import math
from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, functional, image_data_format
from keras_cv_attention_models.attention_layers import (
    activation_by_name,
    batchnorm_with_activation,
    ChannelAffine,
    conv2d_no_bias,
    depthwise_conv2d_no_bias,
    drop_block,
    layer_norm,
    mhsa_with_multi_head_relative_position_embedding,
    MultiHeadRelativePositionalEmbedding,
    se_module,
    output_block,
    window_attention,
    add_pre_post_process,
)
from keras_cv_attention_models.download_and_load import reload_model_weights

PRETRAINED_DICT = {
    "maxvit_tiny": {"imagenet": {224: "e5cfd6a6bd4dea939860b6d8a29a911a", 384: "25b074ca6258d526c4d931f887599fe1", 512: "c745836c38f0ea98fef30bb3186aaf17"}},
    "maxvit_small": {"imagenet": {224: "6bbaff1c6316486c3ac29b607d9ebb13", 384: "85d2b77eab2f2645277a00b3f35d77d9", 512: "63b6c283ea6930fdcbe9b0e2f0bdbe6f"}},
    "maxvit_base": {
        "imagenet": {224: "00c833043b87ef2861ecf79820d827e0", 384: "1b6941018ed54944083add016780d9bd", 512: "530f48f41d49f3ad19cb2b7a5f790517"},
        "imagenet21k": {224: "9b89418259927f12b59df91bdaecae82"},
        "imagenet21k-ft1k": {384: "9b3f8c948e657450c856b0efb305dc7c", 512: "b0d554b7ea52a8bc87dc6790377f4672"},
    },
    "maxvit_large": {
        "imagenet": {224: "93d079fa8171986cc272f6fb4e9b0255", 384: "507afb8b65a99c4dd3c915e3525bdd03", 512: "8ac514a0e3534a24c05eb4087b75454b"},
        "imagenet21k": {224: "a3540387e39efaee5927e7fd9288cc0b"},
        "imagenet21k-ft1k": {384: "4063ab86cf515e7b0ab568cbc5a0726f", 512: "972a7c1cceacb8c1761338c5b6b72605"},
    },
    "maxvit_xlarge": {
        "imagenet21k": {224: "57574f4411cabff061b8946fb768a5df"},
        "imagenet21k-ft1k": {384: "6d78913bc04f716ab029bc6dbc0d9799", 512: "f79fc50827d10f43333321ab03fa563c"},
    },
}


def res_MBConv(inputs, output_channel, conv_short_cut=True, strides=1, expansion=4, se_ratio=0, use_torch_mode=False, drop_rate=0, activation="gelu", name=""):
    if use_torch_mode:
        use_torch_padding, epsilon, momentum = True, 1e-5, 0.9
    else:
        use_torch_padding, epsilon, momentum = False, 0.001, 0.99

    if strides > 1:
        shortcut = layers.AvgPool2D(strides, strides=strides, padding="SAME", name=name + "shortcut_pool")(inputs)
        shortcut = conv2d_no_bias(shortcut, output_channel, 1, strides=1, use_bias=True, name=name + "shortcut_") if conv_short_cut else shortcut
    else:
        shortcut = inputs

    # MBConv
    preact = batchnorm_with_activation(inputs, activation=None, zero_gamma=False, epsilon=epsilon, momentum=momentum, name=name + "preact_")
    nn = conv2d_no_bias(preact, output_channel * expansion, 1, strides=1, padding="same", name=name + "expand_")
    nn = batchnorm_with_activation(nn, activation=activation, epsilon=epsilon, momentum=momentum, name=name + "expand_")
    nn = depthwise_conv2d_no_bias(nn, 3, strides=strides, padding="SAME", use_torch_padding=use_torch_padding, name=name + "MB_")
    nn = batchnorm_with_activation(nn, activation=activation, zero_gamma=False, epsilon=epsilon, momentum=momentum, name=name + "MB_dw_")
    if se_ratio:
        nn = se_module(nn, se_ratio=se_ratio / expansion, activation="swish", name=name + "se/")
    nn = conv2d_no_bias(nn, output_channel, 1, strides=1, use_bias=True, padding="same", name=name + "MB_pw_")
    nn = drop_block(nn, drop_rate=drop_rate, name=name)
    # print(f"{shortcut.shape = }, {nn.shape = }, {strides = }")
    return layers.Add(name=name + "output")([shortcut, nn])


def res_attn_ffn(inputs, output_channel, head_dimension=32, window_size=7, expansion=4, is_grid=False, drop_rate=0, layer_scale=0, activation="gelu", name=""):
    input_channel = inputs.shape[-1]  # Channels_last only
    attn = layer_norm(inputs, axis=-1, name=name + "attn_preact_")
    num_heads = attn.shape[-1] // head_dimension
    # print(f"{inputs.shape = }, {num_heads = }, {window_size = }, {is_grid = }")
    attention_block = lambda inputs, num_heads, name: mhsa_with_multi_head_relative_position_embedding(
        inputs, num_heads=num_heads, qkv_bias=True, out_bias=True, data_format="channels_last", name=name
    )
    attn = window_attention(attn, window_size=window_size, num_heads=num_heads, is_grid=is_grid, attention_block=attention_block, name=name + "window_mhsa/")
    attn = ChannelAffine(use_bias=False, weight_init_value=layer_scale, axis=-1, name=name + "1_gamma")(attn) if layer_scale >= 0 else attn
    attn = drop_block(attn, drop_rate=drop_rate, name=name + "attn_")
    # print(f"{name = }, {inputs.shape = }, {inputs.shape = }, {attn.shape = }")
    attn = layers.Add(name=name + "attn_output")([inputs, attn])

    ffn = layer_norm(attn, axis=-1, name=name + "ffn_preact_")
    ffn = layers.Dense(input_channel * expansion, name=name + "ffn/1_dense")(ffn)
    ffn = activation_by_name(ffn, activation=activation, name=name)
    ffn = layers.Dense(input_channel, name=name + "ffn/2_dense")(ffn)
    ffn = ChannelAffine(use_bias=False, weight_init_value=layer_scale, axis=-1, name=name + "2_gamma")(ffn) if layer_scale >= 0 else ffn
    ffn = drop_block(ffn, drop_rate=drop_rate, name=name + "ffn_")
    return layers.Add(name=name + "ffn_output")([attn, ffn])


def MaxViT(
    num_blocks,
    out_channels,
    stem_width=64,
    strides=[2, 2, 2, 2],
    expansion=4,
    se_ratio=0.25,
    head_dimension=32,
    window_ratio=32,
    output_filter=-1,  # -1 for out_channels[-1], 0 to disable
    use_torch_mode=False,
    layer_scale=-1,
    input_shape=(224, 224, 3),
    num_classes=1000,
    activation="gelu/app",  # means tf.nn.gelu(approximate=True)
    drop_connect_rate=0,
    classifier_activation="softmax",
    dropout=0,
    pretrained=None,
    model_name="maxvit",
    kwargs=None,
):
    # Regard input_shape as force using original shape if first element is None or -1,
    # else assume channel dimention is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)
    inputs = layers.Input(input_shape)
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    height, width = inputs.shape[1:-1] if image_data_format() == "channels_last" else inputs.shape[2:]
    if use_torch_mode:
        use_torch_padding, epsilon, momentum = True, 1e-5, 0.9
    else:
        use_torch_padding, epsilon, momentum = False, 0.001, 0.99

    """ Stem """
    nn = conv2d_no_bias(inputs, stem_width, 3, strides=2, use_bias=True, padding="same", use_torch_padding=use_torch_padding, name="stem_1_")
    nn = batchnorm_with_activation(nn, activation=activation, epsilon=epsilon, momentum=momentum, name="stem_1_")
    nn = conv2d_no_bias(nn, stem_width, 3, strides=1, use_bias=True, padding="same", use_torch_padding=use_torch_padding, name="stem_2_")
    window_size = [int(math.ceil(height / window_ratio)), int(math.ceil(width / window_ratio))]
    # print(f"{window_size = }")

    attn_ffn_common_kwargs = {
        "head_dimension": head_dimension,
        "window_size": window_size,
        "expansion": expansion,
        "layer_scale": layer_scale,
        "activation": activation,
    }

    """ Backbone [1, 2, 3, 4] """
    total_blocks = sum(num_blocks)
    global_block_id = 0
    for stack_id, (num_block, out_channel) in enumerate(zip(num_blocks, out_channels)):
        stack_se_ratio = se_ratio[stack_id] if isinstance(se_ratio, (list, tuple)) else se_ratio
        stack_strides = strides[stack_id] if isinstance(strides, (list, tuple)) else strides
        for block_id in range(num_block):
            name = "stack_{}_block_{}/".format(stack_id + 1, block_id + 1)
            stride = stack_strides if block_id == 0 else 1
            conv_short_cut = True if block_id == 0 and nn.shape[channel_axis] != out_channel else False
            block_se_ratio = stack_se_ratio[block_id] if isinstance(stack_se_ratio, (list, tuple)) else stack_se_ratio
            block_drop_rate = drop_connect_rate * global_block_id / total_blocks
            global_block_id += 1
            nn = res_MBConv(
                nn, out_channel, conv_short_cut, stride, expansion, block_se_ratio, use_torch_mode, block_drop_rate, activation, name=name + "mbconv/"
            )
            nn = nn if image_data_format() == "channels_last" else layers.Permute([2, 3, 1])(nn)  # channels_first -> channels_last
            nn = res_attn_ffn(nn, out_channel, is_grid=False, drop_rate=block_drop_rate, name=name + "block_", **attn_ffn_common_kwargs)
            nn = res_attn_ffn(nn, out_channel, is_grid=True, drop_rate=block_drop_rate, name=name + "grid_", **attn_ffn_common_kwargs)
            nn = nn if image_data_format() == "channels_last" else layers.Permute([3, 1, 2])(nn)  # channels_last -> channels_first

    if num_classes > 0:
        nn = layers.GlobalAveragePooling2D(name="avg_pool")(nn)
        nn = layer_norm(nn, name="post_")
        output_filter = out_channels[-1] if output_filter == -1 else output_filter
        if output_filter > 0:
            nn = layers.Dense(output_filter, name="features")(nn)
            nn = activation_by_name(nn, "tanh", name="features_")
        if dropout > 0:
            nn = layers.Dropout(dropout, name="head_drop")(nn)
        nn = layers.Dense(num_classes, dtype="float32", activation=classifier_activation, name="predictions")(nn)

    model = models.Model(inputs, nn, name=model_name)
    rescale_mode = "tf" if pretrained is not None and pretrained.startswith("imagenet21k") else "torch"  # For testing only
    add_pre_post_process(model, rescale_mode=rescale_mode)
    reload_model_weights(model, PRETRAINED_DICT, "maxvit", pretrained, MultiHeadRelativePositionalEmbedding)
    return model


def MaxViT_Tiny(input_shape=(224, 224, 3), num_classes=1000, drop_connect_rate=0, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [2, 2, 5, 2]
    out_channels = [64, 128, 256, 512]
    stem_width = 64
    return MaxViT(**locals(), model_name="maxvit_tiny", **kwargs)


def MaxViT_Small(input_shape=(224, 224, 3), num_classes=1000, drop_connect_rate=0, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [2, 2, 5, 2]
    out_channels = [96, 192, 384, 768]
    stem_width = 64
    return MaxViT(**locals(), model_name="maxvit_small", **kwargs)


def MaxViT_Base(input_shape=(224, 224, 3), num_classes=1000, drop_connect_rate=0, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [2, 6, 14, 2]
    out_channels = [96, 192, 384, 768]
    stem_width = 64
    return MaxViT(**locals(), model_name="maxvit_base", **kwargs)


def MaxViT_Large(input_shape=(224, 224, 3), num_classes=1000, drop_connect_rate=0, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [2, 6, 14, 2]
    out_channels = [128, 256, 512, 1024]
    stem_width = 128
    return MaxViT(**locals(), model_name="maxvit_large", **kwargs)


def MaxViT_XLarge(input_shape=(224, 224, 3), num_classes=1000, drop_connect_rate=0, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [2, 6, 14, 2]
    out_channels = [192, 384, 768, 1536]
    stem_width = 192
    return MaxViT(**locals(), model_name="maxvit_xlarge", **kwargs)



from keras_cv_attention_models.davit.davit import (
    DaViT,
    DaViT_T,
    DaViT_S,
    DaViT_B,
    DaViT_L,
    DaViT_H,
    DaViT_G,
    multi_head_self_attention_channel,
    window_attention,
)

__head_doc__ = """
Keras implementation of [Github dingmyu/davit](https://github.com/dingmyu/davit).
Paper  [PDF 2204.03645 DaViT: Dual Attention Vision Transformers](https://arxiv.org/pdf/2204.03645.pdf).
"""

__tail_doc__ = """  window_ratio: window_size ratio, window_size = [input_shape[0] // window_ratio, input_shape[1] // window_ratio].
  mlp_ratio: expand ratio for mlp blocks hidden channel.
  stem_patch_size: stem patch size for stem strides.
  layer_scale: layer scale init value. Default `-1` means not applying, any value `>=0` will add a scale value for each block output.
      [Going deeper with Image Transformers](https://arxiv.org/pdf/2103.17239.pdf).
  input_shape: it should have exactly 3 inputs channels, like `(224, 224, 3)`.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  drop_connect_rate: is used for [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382).
      Can be value like `0.2`, indicates the drop probability linearly changes from `0 --> 0.2` for `top --> bottom` layers.
      A higher value means a higher probability will drop the deep branch.
      or `0` to disable (default).
  dropout: dropout rate if top layers is included.
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
  pretrained: None or one of ["imagenet", "token_label"].

Returns:
    A `keras.Model` instance.
"""

DaViT.__doc__ = __head_doc__ + """
Args:
  num_blocks: number of blocks in each stack.
  out_channels: output channels for each stack.
  num_heads: num heads for each stack.
  stem_width: output dimension for stem block, default `-1` for using out_channels[0]
  model_name: string, model name.
""" + __tail_doc__ + """
Model architectures:
  | Model         | Params | FLOPs  | Input | Top1 Acc |
  | ------------- | ------ | ------ | ----- | -------- |
  | DaViT_T       | 28.36M | 4.56G  | 224   | 82.8     |
  | DaViT_S       | 49.75M | 8.83G  | 224   | 84.2     |
  | DaViT_B       | 87.95M | 15.55G | 224   | 84.6     |
  | DaViT_L, 21k  | 196.8M | 103.2G | 384   | 87.5     |
  | DaViT_H, 1.5B | 348.9M | 327.3G | 512   | 90.2     |
  | DaViT_G, 1.5B | 1.406B | 1.022T | 512   | 90.4     |
"""

DaViT_T.__doc__ = __head_doc__ + """
Args:
""" + __tail_doc__

DaViT_S.__doc__ = DaViT_T.__doc__
DaViT_B.__doc__ = DaViT_T.__doc__
DaViT_L.__doc__ = DaViT_T.__doc__
DaViT_H.__doc__ = DaViT_T.__doc__
DaViT_G.__doc__ = DaViT_T.__doc__

multi_head_self_attention_channel.__doc__ = __head_doc__ + """
Multi head self attention on channel dimension. Defined as function, not layer.
It's different from traditional MHSA, that using `attention_scores` shape `[batch, num_heads, key_dim, key_dim]`,
while traditional MHSA `attention_scores` shape `[batch, num_heads, hh * ww, hh * ww]`

Args:
  inputs: input tensor.
  num_heads: Number of attention heads.
  key_dim: Size of each attention head for query, key and value. Default `0` for `key_dim = inputs.shape[-1] // num_heads`.
  out_shape: The expected shape of an output tensor. If not specified, projects back to the input dim.
  out_weight: Boolean, whether use an ouput dense.
  qkv_bias: Boolean, whether the qkv dense layer use bias vectors/matrices.
  out_bias: Boolean, whether the ouput dense layer use bias vectors/matrices.
  attn_dropout: Dropout probability for attention.

Examples:

>>> from keras_cv_attention_models import attention_layers
>>> inputs = keras.layers.Input([14, 16, 256])
>>> nn = attention_layers.multi_head_self_attention_channel(inputs, num_heads=4, out_shape=512, name="attn_")
>>> print(f"{nn.shape = }")
# nn.shape = TensorShape([None, 14, 16, 512])

>>> mm = keras.models.Model(inputs, nn)
>>> mm.summary()
>>> print({ii.name: ii.shape for ii in mm.weights})
# {'attn_qkv/kernel:0': TensorShape([256, 768]), 'attn_output/kernel:0': TensorShape([256, 512])}
"""

window_attention.__doc__ = __head_doc__ + """
Window multi head self attention. Defined as function, not layer.
Typical MHSA with `window_partition` process ahead and `window_reverse` process after.
Also works like a wrapper, perform `window_partition -> attention_block -> window_reverse`.

Args:
  inputs: input tensor.
  window_size: window partition size.
  num_heads: Number of attention heads.
  attention_block: specific callable attention block, instead of preset MHSA.
      3 parameters are required supporting: `inputs, num_heads, name`
  kwargs: any additional kwargs for `attention_block`.

Examples:

>>> from keras_cv_attention_models import attention_layers
>>> inputs = keras.layers.Input([14, 16, 256])
>>> nn = attention_layers.window_attention(inputs, 7, num_heads=4, name="attn_")
>>> print(f"{nn.shape = }")
# nn.shape = TensorShape([None, 14, 16, 256])

>>> mm = keras.models.Model(inputs, nn)
>>> mm.summary()
>>> print({ii.name: ii.shape for ii in mm.weights})
# {'attn_qkv/kernel:0': TensorShape([256, 768]),
#  'attn_qkv/bias:0': TensorShape([768]),
#  'attn_output/kernel:0': TensorShape([256, 256]),
#  'attn_output/bias:0': TensorShape([256])}
"""



import os
import json
import numpy as np


CLASS_INDEX_IMAGENET = None
CLASS_INDEX_IMAGENET21K = None

""" Imagenet evaluation """


class TorchModelInterf:
    def __init__(self, model):
        import torch

        self.torch = torch
        cvd = os.environ.get("CUDA_VISIBLE_DEVICES", "").strip()
        device_name = "cuda:0" if len(cvd) > 0 and int(cvd) != -1 else "cpu"
        self.device = self.torch.device(device_name)
        self.model = model.cuda(device_name)

    def __call__(self, imgs):
        # print(imgs.shape, imgs[0])
        imgs = imgs.numpy() if hasattr(imgs, "numpy") else imgs
        output = self.model(self.torch.from_numpy(imgs).permute([0, 3, 1, 2]).to(self.device).float())
        return output.cpu().detach().numpy()


class TFLiteModelInterf:
    def __init__(self, model_path):
        import tensorflow as tf

        self.tf = tf
        self.interpreter = tf.lite.Interpreter(model_path=model_path)
        self.name = os.path.splitext(os.path.basename(model_path))[0]

        input_details = self.interpreter.get_input_details()[0]
        self.input_dtype = input_details["dtype"]
        self.input_index = input_details["index"]
        self.input_shape = input_details["shape"].tolist()

        output_details = self.interpreter.get_output_details()[0]
        self.output_dtype = output_details["dtype"]
        self.output_index = output_details["index"]
        self.output_shape = output_details["shape"].tolist()

        if self.input_dtype == tf.uint8 or self.output_dtype == tf.uint8:
            self.input_scale, self.input_zero_point = input_details.get("quantization", (1.0, 0.0))
            self.output_scale, self.output_zero_point = output_details.get("quantization", (1.0, 0.0))
            self.__interf__ = self.__uint8_interf__
        else:
            self.__interf__ = self.__float_interf__

        self.interpreter.allocate_tensors()

    def __float_interf__(self, img):
        img = self.tf.cast(img, self.input_dtype)
        self.interpreter.set_tensor(self.input_index, self.tf.expand_dims(img, 0))
        self.interpreter.invoke()
        return self.interpreter.get_tensor(self.output_index).copy()

    def __uint8_interf__(self, img):
        img = self.tf.cast(img / self.input_scale + self.input_zero_point, self.input_dtype)
        self.interpreter.set_tensor(self.input_index, self.tf.expand_dims(img, 0))
        self.interpreter.invoke()
        pred = self.interpreter.get_tensor(self.output_index).copy()
        return (pred.astype("float32") - self.output_zero_point) * self.output_scale

    def __call__(self, imgs):
        # print(imgs.shape, imgs[0])
        preds = []
        for img in imgs:
            pred = self.__interf__(img)
            preds.append(pred)
        return self.tf.concat(preds, 0).numpy()


class ONNXModelInterf:
    def __init__(self, model_file):
        import onnx
        import onnxruntime as ort

        ort.set_default_logger_severity(3)
        model_content = onnx.load_model(model_file)
        model_content.graph.input[0].type.tensor_type.shape.dim[0].dim_param = "None"  # Set batch_size as dynamic
        model_content.graph.output[0].type.tensor_type.shape.dim[0].dim_param = "None"  # Set batch_size as dynamic

        self.ort_session = ort.InferenceSession(model_content.SerializeToString())
        self.input_name = self.ort_session.get_inputs()[0].name
        self.output_names = [self.ort_session.get_outputs()[0].name]
        self.input_shape = self.ort_session.get_inputs()[0].shape
        self.output_shape = self.ort_session.get_outputs()[0].shape

        # Regard channel shape as the smallest value, and if it's on dimension 1, regard as channel first NCHW format
        self.data_format = "NCHW" if min(self.input_shape[1:]) == self.input_shape[1] else "NHWC"

    def __call__(self, imgs):
        imgs = imgs.numpy() if hasattr(imgs, "numpy") else imgs
        if self.data_format == "NCHW":
            imgs = imgs.transpose(0, 3, 1, 2)
        return self.ort_session.run(self.output_names, {self.input_name: imgs.astype("float32")})[0]
        # return np.array([self.ort_session.run(self.output_names, {self.input_name: imgs[None]})[0][0] for img in imgs])


def evaluation(
    model, data_name="imagenet2012", input_shape=None, batch_size=64, central_crop=1.0, resize_method="bicubic", antialias=True, rescale_mode="auto"
):
    import types
    from tqdm import tqdm
    from keras_cv_attention_models.backend import models
    from keras_cv_attention_models.model_surgery import change_model_input_shape
    from keras_cv_attention_models.imagenet import data

    if isinstance(model, models.Model):
        input_shape = model.input_shape[1:-1] if input_shape is None else input_shape[:2]
        model_interf = change_model_input_shape(model, input_shape)
        print(">>>> Using input_shape {} for Keras model.".format(input_shape))
    elif isinstance(model, TFLiteModelInterf) or (isinstance(model, str) and model.endswith(".tflite")):
        model_interf = model if isinstance(model, TFLiteModelInterf) else TFLiteModelInterf(model)
        input_shape = model_interf.input_shape[1:-1]
        print(">>>> Using input_shape {} for TFLite model.".format(input_shape))
    elif isinstance(model, types.LambdaType):
        model_interf = model
    elif isinstance(model, str) and model.endswith(".onnx"):
        model_interf = ONNXModelInterf(model)
        input_shape = model_interf.input_shape[2:] if model_interf.data_format == "NCHW" else model_interf.input_shape[1:-1]
        print(">>>> Using input_shape {} for ONNX model, data_format: {}.".format(input_shape, model_interf.data_format))
    else:
        model_interf = TorchModelInterf(model)
        assert input_shape is not None

    if isinstance(rescale_mode, str) and rescale_mode.lower() == "auto":
        rescale_mode = getattr(model, "rescale_mode", "torch")
        print(">>>> rescale_mode:", rescale_mode)

    test_dataset = data.init_dataset(
        data_name,
        input_shape=input_shape,
        batch_size=batch_size,
        eval_central_crop=central_crop,
        resize_method=resize_method,
        resize_antialias=antialias,
        rescale_mode=rescale_mode,
    )[1]

    y_true, y_pred_top_1, y_pred_top_5 = [], [], []
    for img_batch, true_labels in tqdm(test_dataset.as_numpy_iterator(), "Evaluating", total=len(test_dataset)):
        predicts = model_interf(img_batch)
        predicts = np.array(predicts.detach() if hasattr(predicts, "detach") else predicts)
        pred_argsort = predicts.argsort(-1)
        y_pred_top_1.extend(pred_argsort[:, -1])
        y_pred_top_5.extend(pred_argsort[:, -5:])
        y_true.extend(np.array(true_labels).argmax(-1))
    y_true, y_pred_top_1, y_pred_top_5 = np.array(y_true), np.array(y_pred_top_1), np.array(y_pred_top_5)
    accuracy_1 = np.sum(y_true == y_pred_top_1) / y_true.shape[0]
    accuracy_5 = np.sum([ii in jj for ii, jj in zip(y_true, y_pred_top_5)]) / y_true.shape[0]
    print(">>>> Accuracy top1:", accuracy_1, "top5:", accuracy_5)
    return y_true, y_pred_top_1, y_pred_top_5


""" Decode predictions """


def decode_predictions(preds, top=5):
    """Similar function from keras.applications.imagenet_utils.decode_predictions, just also supporting imagenet21k class index"""
    from keras_cv_attention_models.backend import get_file

    preds = np.array(preds.detach() if hasattr(preds, "detach") else preds)
    if len(preds.shape) != 2 or (preds.shape[-1] != 1000 and preds.shape[-1] != 21843):
        print("[Error] not imagenet or imagenet21k prediction, not supported")
        return

    is_imagenet = preds.shape[-1] == 1000

    global CLASS_INDEX_IMAGENET
    global CLASS_INDEX_IMAGENET21K
    if (is_imagenet and CLASS_INDEX_IMAGENET is None) or CLASS_INDEX_IMAGENET21K is None:
        if is_imagenet:
            url = "https://github.com/leondgarse/keras_cv_attention_models/releases/download/efficientdet/imagenet_class_index.json"
        else:
            url = "https://github.com/leondgarse/keras_cv_attention_models/releases/download/efficientdet/imagenet21k_class_index.json"

        class_index_path = get_file(origin=url)
        with open(class_index_path) as ff:
            if is_imagenet:
                CLASS_INDEX_IMAGENET = json.load(ff)
            else:
                CLASS_INDEX_IMAGENET21K = json.load(ff)
    class_index = CLASS_INDEX_IMAGENET if is_imagenet else CLASS_INDEX_IMAGENET21K

    results = []
    for pred in preds:
        top_indices = pred.argsort()[-top:][::-1]
        result = [tuple(class_index[str(i)]) + (pred[i],) for i in top_indices]
        result.sort(key=lambda x: x[2], reverse=True)
        results.append(result)
    return results


def decode_predictions_imagenet21k(preds, top=5):
    return decode_predictions(preds, top=top)  # just keeping for compiling with ealier


""" Plotting related """


def parse_timm_log(log_file, pick_keys=None):
    with open(log_file, "r") as ff:
        aa = ff.readlines()

    """ Find pattern for train epoch end """
    train_epoch_started, train_epoch_end_pattern, previous_line = False, "", ""
    for ii in aa:
        if ii.startswith("Train:"):
            train_epoch_started = True
            previous_line = ii
        elif train_epoch_started and ii.startswith("Test:"):
            train_epoch_end_pattern = previous_line.split("[")[1].split("]")[0].strip()
            break

    """ Find pattern for test end """
    test_epoch_started, test_epoch_end_pattern, previous_line = False, "", ""
    for ii in aa:
        if ii.startswith("Test:"):
            test_epoch_started = True
            previous_line = ii
        elif test_epoch_started and not ii.startswith("Train:"):
            test_epoch_end_pattern = previous_line.split("[")[1].split("]")[0].strip()
            break
    print("train_epoch_end_pattern = {}, test_epoch_end_pattern = {}".format(train_epoch_end_pattern, test_epoch_end_pattern))

    split_func = lambda xx, ss, ee: float(xx.split(ss)[1].strip().split(ee)[0].split("(")[-1].split(")")[0])
    train_loss = [split_func(ii, "Loss:", "Time:") for ii in aa if train_epoch_end_pattern in ii]
    lr = [split_func(ii, "LR:", "Data:") for ii in aa if train_epoch_end_pattern in ii]
    val_loss = [split_func(ii, "Loss:", "Acc@1:") for ii in aa if test_epoch_end_pattern in ii]
    val_acc = [split_func(ii, "Acc@1:", "Acc@5:") for ii in aa if test_epoch_end_pattern in ii]
    if val_acc[-1] > 1:
        val_acc = [ii / 100.0 for ii in val_acc]

    # print(f"{len(train_loss) = }, {len(lr) = }, {len(val_loss) = }, {len(val_acc) = }")
    hh = {"loss": train_loss, "lr": lr, "val_loss": val_loss, "val_acc": val_acc}
    return hh if pick_keys is None else {kk: hh[kk] for kk in pick_keys}


def combine_hist_into_one(hist_list, save_file=None):
    hh = {}
    for hist in hist_list:
        with open(hist, "r") as ff:
            aa = json.load(ff)
        for kk, vv in aa.items():
            hh.setdefault(kk, []).extend(vv)

    if save_file:
        with open(save_file, "w") as ff:
            json.dump(hh, ff)
    return hh


def curve_fit(source, target_len=10, skip=5, use_recent=40):
    from scipy.optimize import curve_fit

    def func_curv(x, a, b, c, d):
        pp = np.log(x)
        # pp = 1 / x
        return a * pp**3 + b * pp**2 + c * pp + d

    recent_source = source[skip:]
    use_recent = len(source) if use_recent == -1 else use_recent
    if len(recent_source) > use_recent:
        recent_source = recent_source[-use_recent:]
    start_pos = len(source) - len(recent_source)
    popt, pcov = curve_fit(func_curv, np.arange(start_pos, len(source)), recent_source)
    return list(source[: -len(recent_source)]) + func_curv(np.arange(start_pos, len(source) + target_len), *popt).tolist()


def plot_and_peak_scatter(ax, source_array, peak_method, label, skip_first=0, color=None, va="bottom", pred_curve=0, **kwargs):
    array = source_array[skip_first:]
    for id, ii in enumerate(array):
        if np.isnan(ii):
            array[id] = array[id - 1]
    ax.plot(range(skip_first, skip_first + len(array)), array, label=label, color=color, **kwargs)
    color = ax.lines[-1].get_color() if color is None else color
    pp = peak_method(array)
    vv = array[pp]
    ax.scatter(pp + skip_first, vv, color=color, marker="v")
    # ax.text(pp + skip_first, vv, "{:.4f}".format(vv), va=va, ha="right", color=color, fontsize=fontsize, rotation=0)
    ax.text(pp + skip_first, vv, "{:.4f}".format(vv), va=va, ha="right", color=color, rotation=0)

    if pred_curve > 0:
        kwargs.pop("linestyle", None)
        pred_array = curve_fit(source_array, pred_curve)[skip_first:]
        ax.plot(range(skip_first, skip_first + len(pred_array)), pred_array, color=color, linestyle=":", **kwargs)


def plot_hists(hists, names=None, base_size=6, addition_plots=["lr"], text_va=["bottom"], skip_first=0, pred_curve=0):
    import matplotlib as mpl
    import matplotlib.pyplot as plt

    mpl.rcParams["lines.linewidth"] = base_size / 8 * 1.5  # 8 -> 1.5
    mpl.rcParams["font.size"] = base_size * 2 - 4  # 8 -> 12, 6 -> 8
    mpl.rcParams["legend.fontsize"] = base_size * 2 - 4
    mpl.rcParams["xtick.labelsize"] = base_size * 2 - 4
    mpl.rcParams["ytick.labelsize"] = base_size * 2 - 4
    mpl.rcParams["xtick.major.pad"] = base_size / 2 - 2  # 8 -> 4, 6 -> 3
    mpl.rcParams["ytick.major.pad"] = base_size / 2 - 2

    num_axes = (2 + len(addition_plots)) if addition_plots is not None and len(addition_plots) != 0 else 2
    fig, axes = plt.subplots(1, num_axes, figsize=(num_axes * base_size, base_size))
    hists = [hists] if isinstance(hists, (str, dict)) else hists
    names = names if isinstance(names, (list, tuple)) else [names]
    for id, hist in enumerate(hists):
        name = names[min(id, len(names) - 1)] if names != None else None
        cur_va = text_va[id % len(text_va)]
        if isinstance(hist, str):
            name = name if name != None else os.path.splitext(os.path.basename(hist))[0]
            with open(hist, "r") as ff:
                hist = json.load(ff)
        name = name if name != None else str(id)

        acc_key = "acc"
        if acc_key not in hist:
            all_acc_key = [ii for ii in hist.keys() if "acc" in ii and "val" not in ii]
            acc_key = "acc" if len(all_acc_key) == 0 else all_acc_key[0]

        val_acc_key = "val_acc"
        if "val_ap_ar" in hist:  # It's coco hist
            hist["val_ap_0.50:0.95"] = [ii[0] for ii in hist["val_ap_ar"]]
            val_acc_key = "val_ap 0.50:0.95"
        elif val_acc_key not in hist:
            all_val_acc_key = [ii for ii in hist.keys() if "acc" in ii and "val" in ii]
            val_acc_key = "val_acc" if len(all_val_acc_key) == 0 else all_val_acc_key[0]

        cur_pred_curve = pred_curve[min(id, len(pred_curve) - 1)] if isinstance(pred_curve, (list, tuple)) else pred_curve
        plot_and_peak_scatter(axes[0], hist["loss"], np.argmin, name + " loss", skip_first, None, cur_va, pred_curve=cur_pred_curve)
        color = axes[0].lines[-1].get_color()
        val_loss = hist.get("val_loss", [])
        if len(val_loss) > 0 and "val_loss" not in addition_plots:
            plot_and_peak_scatter(axes[0], val_loss, np.argmin, name + " val_loss", skip_first, color, cur_va, cur_pred_curve, linestyle="--")

        acc = hist.get(acc_key, [])
        if len(acc) > 0:  # For timm log
            plot_and_peak_scatter(axes[1], acc, np.argmax, name + " " + acc_key, skip_first, color, cur_va, cur_pred_curve)

        val_acc = hist.get(val_acc_key, [])
        if len(val_acc) > 0:  # For timm log
            plot_and_peak_scatter(axes[1], val_acc, np.argmax, name + " " + val_acc_key, skip_first, color, cur_va, cur_pred_curve, linestyle="--")

        if addition_plots is not None and len(addition_plots) != 0:
            for id, ii in enumerate(addition_plots):
                if len(hist.get(ii, [])) > 0:
                    peak_method = np.argmin if "loss" in ii else np.argmax
                    plot_and_peak_scatter(axes[2 + id], hist[ii], peak_method, name + " " + ii, skip_first, color, cur_va, cur_pred_curve)
    for ax in axes:
        ax.legend()
        ax.grid(True)
    fig.tight_layout()
    fig.tight_layout()  # again
    plt.show()
    return fig



import numpy as np
from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, functional, initializers, image_data_format

BATCH_NORM_DECAY = 0.9
BATCH_NORM_EPSILON = 1e-5
TF_BATCH_NORM_EPSILON = 0.001
LAYER_NORM_EPSILON = 1e-5
CONV_KERNEL_INITIALIZER = "glorot_uniform"


""" Wrapper for default parameters """


@backend.register_keras_serializable(package="kecamCommon")
def hard_swish(inputs):
    """`out = xx * relu6(xx + 3) / 6`, arxiv: https://arxiv.org/abs/1905.02244"""
    return inputs * functional.relu6(inputs + 3) / 6


@backend.register_keras_serializable(package="kecamCommon")
def hard_sigmoid_torch(inputs):
    """https://pytorch.org/docs/stable/generated/torch.nn.Hardsigmoid.html
    toch.nn.Hardsigmoid: 0 if x <= −3 else (1 if x >= 3 else x / 6 + 1/2)
    keras.activations.hard_sigmoid: 0 if x <= −2.5 else (1 if x >= 2.5 else x / 5 + 1/2) -> tf.clip_by_value(inputs / 5 + 0.5, 0, 1)
    """
    return functional.clip_by_value(inputs / 6 + 0.5, 0, 1)


@backend.register_keras_serializable(package="kecamCommon")
def mish(inputs):
    """Mish: A Self Regularized Non-Monotonic Neural Activation Function.
    Paper: [Mish: A Self Regularized Non-Monotonic Neural Activation Function](https://arxiv.org/abs/1908.08681)
    Copied from https://github.com/tensorflow/addons/blob/master/tensorflow_addons/activations/mish.py
    """
    return inputs * functional.tanh(functional.softplus(inputs))


@backend.register_keras_serializable(package="kecamCommon")
def phish(inputs):
    """Phish is defined as f(x) = xTanH(GELU(x)) with no discontinuities in the f(x) derivative.
    Paper: https://www.techrxiv.org/articles/preprint/Phish_A_Novel_Hyper-Optimizable_Activation_Function/17283824
    """
    return inputs * functional.tanh(functional.gelu(inputs))


def gelu_linear(inputs):
    """
    >>> from keras_cv_attention_models.common_layers import gelu_linear
    >>> xx = np.arange(-4, 4, 0.01)
    >>> plt.plot(xx, tf.nn.gelu(xx), label='gelu')
    >>> plt.plot(xx, tf.nn.gelu(xx, approximate=True), label='gelu, approximate')
    >>> plt.plot(xx, gelu_linear(xx), label='gelu_linear')
    >>> plt.legend()
    >>> plt.grid(True)
    """
    inputs_abs = functional.abs(inputs)
    inputs_sign = functional.sign(inputs)

    erf = inputs_abs * -0.7071
    erf = functional.relu(erf + 1.769)
    erf = erf ** 2 * -0.1444 + 0.5
    return inputs * (erf * inputs_sign + 0.5)


def activation_by_name(inputs, activation="relu", name=None):
    """Typical Activation layer added hard_swish and prelu."""
    if activation is None:
        return inputs

    layer_name = name and activation and name + activation
    activation_lower = activation.lower()
    if activation_lower == "hard_swish":
        return layers.Activation(activation=hard_swish, name=layer_name)(inputs)
    elif activation_lower == "mish":
        return layers.Activation(activation=mish, name=layer_name)(inputs)
    elif activation_lower == "phish":
        return layers.Activation(activation=phish, name=layer_name)(inputs)
    elif activation_lower == "prelu":
        shared_axes = list(range(1, len(inputs.shape)))
        shared_axes.pop(-1 if backend.image_data_format() == "channels_last" else 0)
        # print(f"{shared_axes = }")
        return layers.PReLU(shared_axes=shared_axes, alpha_initializer=initializers.Constant(0.25), name=layer_name)(inputs)
    elif activation_lower.startswith("gelu/app"):
        # gelu/approximate
        return functional.gelu(inputs, approximate=True, name=layer_name)
    elif activation_lower.startswith("gelu/linear"):
        return gelu_linear(inputs)
    elif activation_lower.startswith("leaky_relu/"):
        # leaky_relu with alpha parameter
        alpha = float(activation_lower.split("/")[-1])
        return layers.LeakyReLU(alpha=alpha, name=layer_name)(inputs)
    elif activation_lower == ("hard_sigmoid_torch"):
        return layers.Activation(activation=hard_sigmoid_torch, name=layer_name)(inputs)
    elif activation_lower == ("squaredrelu") or activation_lower == ("squared_relu"):
        return functional.pow(functional.relu(inputs), 2)  # Squared ReLU: https://arxiv.org/abs/2109.08668
    elif activation_lower == ("starrelu") or activation_lower == ("star_relu"):
        from keras_cv_attention_models.nfnets.nfnets import ZeroInitGain

        # StarReLU: s * relu(x) ** 2 + b
        return ZeroInitGain(use_bias=True, weight_init_value=1.0, name=layer_name)(functional.pow(functional.relu(inputs), 2))
    else:
        return layers.Activation(activation=activation, name=layer_name)(inputs)


@backend.register_keras_serializable(package="kecamCommon")
class EvoNormalization(layers.Layer):
    def __init__(self, nonlinearity=True, num_groups=-1, zero_gamma=False, momentum=0.99, epsilon=0.001, data_format="auto", **kwargs):
        # [evonorm](https://github.com/tensorflow/tpu/blob/master/models/official/resnet/resnet_model.py)
        # EVONORM_B0: nonlinearity=True, num_groups=-1
        # EVONORM_S0: nonlinearity=True, num_groups > 0
        # EVONORM_B0 / EVONORM_S0 linearity: nonlinearity=False, num_groups=-1
        # EVONORM_S0A linearity: nonlinearity=False, num_groups > 0
        super().__init__(**kwargs)
        self.data_format, self.nonlinearity, self.zero_gamma, self.num_groups = data_format, nonlinearity, zero_gamma, num_groups
        self.momentum, self.epsilon = momentum, epsilon
        self.is_channels_first = (
            True if data_format == "channels_first" or (data_format == "auto" and backend.image_data_format() == "channels_first") else False
        )

    def build(self, input_shape):
        all_axes = list(range(len(input_shape)))
        param_shape = [1] * len(input_shape)
        if self.is_channels_first:
            param_shape[1] = input_shape[1]
            self.reduction_axes = all_axes[:1] + all_axes[2:]
        else:
            param_shape[-1] = input_shape[-1]
            self.reduction_axes = all_axes[:-1]

        self.gamma = self.add_weight(name="gamma", shape=param_shape, initializer="zeros" if self.zero_gamma else "ones", trainable=True)
        self.beta = self.add_weight(name="beta", shape=param_shape, initializer="zeros", trainable=True)
        if self.num_groups <= 0:  # EVONORM_B0
            self.moving_variance = self.add_weight(
                name="moving_variance",
                shape=param_shape,
                initializer="ones",
                # synchronization=tf.VariableSynchronization.ON_READ,
                trainable=False,
                # aggregation=tf.VariableAggregation.MEAN,
            )
        if self.nonlinearity:
            self.vv = self.add_weight(name="vv", shape=param_shape, initializer="ones", trainable=True)

        if self.num_groups > 0:  # EVONORM_S0
            channels_dim = input_shape[1] if self.is_channels_first else input_shape[-1]
            num_groups = int(self.num_groups)
            while num_groups > 1:
                if channels_dim % num_groups == 0:
                    break
                num_groups -= 1
            self.__num_groups__ = num_groups
            self.groups_dim = channels_dim // self.__num_groups__

            if self.is_channels_first:
                self.group_shape = [-1, self.__num_groups__, self.groups_dim, *input_shape[2:]]
                self.group_reduction_axes = list(range(2, len(self.group_shape)))  # [2, 3, 4]
                self.group_axes = 2
                self.var_shape = [-1, *param_shape[1:]]
            else:
                self.group_shape = [-1, *input_shape[1:-1], self.__num_groups__, self.groups_dim]
                self.group_reduction_axes = list(range(1, len(self.group_shape) - 2)) + [len(self.group_shape) - 1]  # [1, 2, 4]
                self.group_axes = -1
                self.var_shape = [-1, *param_shape[1:]]

    def __group_std__(self, inputs):
        # _group_std, https://github.com/tensorflow/tpu/blob/main/models/official/resnet/resnet_model.py#L171
        grouped = functional.reshape(inputs, self.group_shape)
        _, var = functional.moments(grouped, self.group_reduction_axes, keepdims=True)
        std = functional.sqrt(var + self.epsilon)
        std = functional.repeat(std, self.groups_dim, axis=self.group_axes)
        return functional.reshape(std, self.var_shape)

    def __batch_std__(self, inputs, training=None):
        # _batch_std, https://github.com/tensorflow/tpu/blob/main/models/official/resnet/resnet_model.py#L120
        def _call_train_():
            _, var = functional.moments(inputs, self.reduction_axes, keepdims=True)
            # update_op = tf.assign_sub(moving_variance, (moving_variance - variance) * (1 - decay))
            delta = (self.moving_variance - var) * (1 - self.momentum)
            self.moving_variance.assign_sub(delta)
            return var

        def _call_test_():
            return self.moving_variance

        var = backend.in_train_phase(_call_train_, _call_test_, training=training)
        return functional.sqrt(var + self.epsilon)

    def __instance_std__(self, inputs):
        # _instance_std, https://github.com/tensorflow/tpu/blob/main/models/official/resnet/resnet_model.py#L111
        # axes = [1, 2] if data_format == 'channels_last' else [2, 3]
        _, var = functional.moments(inputs, self.reduction_axes[1:], keepdims=True)
        return functional.sqrt(var + self.epsilon)

    def call(self, inputs, training=None, **kwargs):
        if self.nonlinearity and self.num_groups > 0:  # EVONORM_S0
            den = self.__group_std__(inputs)
            inputs = inputs * functional.sigmoid(self.vv * inputs) / den
        elif self.num_groups > 0:  # EVONORM_S0a
            # EvoNorm2dS0a https://github.com/rwightman/pytorch-image-models/blob/main/timm/models/layers/evo_norm.py#L239
            den = self.__group_std__(inputs)
            inputs = inputs / den
        elif self.nonlinearity:  # EVONORM_B0
            left = self.__batch_std__(inputs, training)
            right = self.vv * inputs + self.__instance_std__(inputs)
            inputs = inputs / functional.maximum(left, right)
        return inputs * self.gamma + self.beta

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "nonlinearity": self.nonlinearity,
                "zero_gamma": self.zero_gamma,
                "num_groups": self.num_groups,
                "momentum": self.momentum,
                "epsilon": self.epsilon,
                "data_format": self.data_format,
            }
        )
        return config


def batchnorm_with_activation(
    inputs, activation=None, zero_gamma=False, epsilon=1e-5, momentum=0.9, axis="auto", act_first=False, use_evo_norm=False, evo_norm_group_size=-1, name=None
):
    """Performs a batch normalization followed by an activation."""
    if use_evo_norm:
        nonlinearity = False if activation is None else True
        num_groups = inputs.shape[-1] // evo_norm_group_size  # Currently using gorup_size as parameter only
        return EvoNormalization(nonlinearity, num_groups=num_groups, zero_gamma=zero_gamma, epsilon=epsilon, momentum=momentum, name=name + "evo_norm")(inputs)

    bn_axis = (-1 if backend.image_data_format() == "channels_last" else 1) if axis == "auto" else axis
    gamma_initializer = initializers.zeros() if zero_gamma else initializers.ones()
    if act_first and activation:
        inputs = activation_by_name(inputs, activation=activation, name=name)
    nn = layers.BatchNormalization(
        axis=bn_axis,
        momentum=momentum,
        epsilon=epsilon,
        gamma_initializer=gamma_initializer,
        name=name and name + "bn",
    )(inputs)
    if not act_first and activation:
        nn = activation_by_name(nn, activation=activation, name=name)
    return nn


def layer_norm(inputs, zero_gamma=False, epsilon=LAYER_NORM_EPSILON, center=True, axis="auto", name=None):
    """Typical LayerNormalization with epsilon=1e-5"""
    norm_axis = (-1 if backend.image_data_format() == "channels_last" else 1) if axis == "auto" else axis
    gamma_init = initializers.zeros() if zero_gamma else initializers.ones()
    return layers.LayerNormalization(axis=norm_axis, epsilon=epsilon, gamma_initializer=gamma_init, center=center, name=name and name + "ln")(inputs)


def group_norm(inputs, groups=32, epsilon=BATCH_NORM_EPSILON, axis="auto", name=None):
    """Typical GroupNormalization with epsilon=1e-5"""
    if hasattr(layers, "GroupNormalization"):
        GroupNormalization = layers.GroupNormalization  # GroupNormalization is added after TF 2.11.0
    else:
        from tensorflow_addons.layers import GroupNormalization

    norm_axis = (-1 if backend.image_data_format() == "channels_last" else 1) if axis == "auto" else axis
    return GroupNormalization(groups=groups, axis=norm_axis, epsilon=epsilon, name=name and name + "group_norm")(inputs)


def conv2d_no_bias(inputs, filters, kernel_size=1, strides=1, padding="VALID", use_bias=False, groups=1, use_torch_padding=True, name=None, **kwargs):
    """Typical Conv2D with `use_bias` default as `False` and fixed padding"""
    kernel_size = kernel_size if isinstance(kernel_size, (list, tuple)) else (kernel_size, kernel_size)
    if isinstance(padding, str):
        padding = padding.upper()
        pad = (kernel_size[0] // 2, kernel_size[1] // 2) if use_torch_padding and padding == "SAME" else (0, 0)
    else:  # int or list or tuple with specific value
        pad = padding if isinstance(padding, (list, tuple)) else (padding, padding)
        padding = "SAME" if max(pad) > 0 else "VALID"

    if use_torch_padding and backend.is_tensorflow_backend and padding == "SAME":
        inputs = layers.ZeroPadding2D(padding=pad, name=name and name + "pad")(inputs)
        padding = "VALID"

    groups = max(1, groups)
    return layers.Conv2D(
        filters,
        kernel_size,
        strides=strides,
        padding="VALID" if padding == "VALID" else (pad if use_torch_padding else "SAME"),
        use_bias=use_bias,
        groups=groups,
        kernel_initializer=CONV_KERNEL_INITIALIZER,
        name=name and name + "conv",
        **kwargs,
    )(inputs)


def depthwise_conv2d_no_bias(inputs, kernel_size, strides=1, padding="VALID", use_bias=False, use_torch_padding=True, name=None, **kwargs):
    """Typical DepthwiseConv2D with `use_bias` default as `False` and fixed padding"""
    kernel_size = kernel_size if isinstance(kernel_size, (list, tuple)) else (kernel_size, kernel_size)
    if isinstance(padding, str):
        padding = padding.upper()
        pad = (kernel_size[0] // 2, kernel_size[1] // 2) if use_torch_padding and padding == "SAME" else (0, 0)
    else:  # int or list or tuple with specific value
        pad = padding if isinstance(padding, (list, tuple)) else (padding, padding)
        padding = "SAME" if max(pad) > 0 else "VALID"

    if use_torch_padding and backend.is_tensorflow_backend and padding == "SAME":
        inputs = layers.ZeroPadding2D(padding=pad, name=name and name + "pad")(inputs)
        padding = "VALID"

    return layers.DepthwiseConv2D(
        kernel_size,
        strides=strides,
        padding="VALID" if padding == "VALID" else (pad if use_torch_padding else "SAME"),
        use_bias=use_bias,
        depthwise_initializer=CONV_KERNEL_INITIALIZER,
        name=name and name + "dw_conv",
        **kwargs,
    )(inputs)


""" Blocks """


def output_block(inputs, filters=0, activation="relu", num_classes=1000, drop_rate=0, classifier_activation="softmax", is_torch_mode=True, act_first=False):
    nn = inputs
    if filters > 0:  # efficientnet like
        bn_eps = BATCH_NORM_EPSILON if is_torch_mode else TF_BATCH_NORM_EPSILON
        nn = conv2d_no_bias(nn, filters, 1, strides=1, use_bias=act_first, use_torch_padding=is_torch_mode, name="features_")  # Also use_bias for act_first
        nn = batchnorm_with_activation(nn, activation=activation, act_first=act_first, epsilon=bn_eps, name="features_")

    if num_classes > 0:
        nn = layers.GlobalAveragePooling2D(name="avg_pool")(nn) if len(nn.shape) == 4 else nn
        if drop_rate > 0:
            nn = layers.Dropout(drop_rate, name="head_drop")(nn)
        nn = layers.Dense(num_classes, dtype="float32", activation=classifier_activation, name="predictions")(nn)
    return nn


def global_context_module(inputs, use_attn=True, ratio=0.25, divisor=1, activation="relu", use_bias=True, name=None):
    """Global Context Attention Block, arxiv: https://arxiv.org/pdf/1904.11492.pdf"""
    is_channels_last = image_data_format() == "channels_last"
    filters = inputs.shape[-1 if is_channels_last else 1]
    height_axis, width_axis = (1, 2) if is_channels_last else (2, 3)
    height, width = inputs.shape[height_axis], inputs.shape[width_axis]

    # activation could be ("relu", "hard_sigmoid")
    hidden_activation, output_activation = activation if isinstance(activation, (list, tuple)) else (activation, "sigmoid")
    reduction = make_divisible(filters * ratio, divisor, limit_round_down=0.0)

    if use_attn:
        attn = layers.Conv2D(1, kernel_size=1, use_bias=use_bias, kernel_initializer=CONV_KERNEL_INITIALIZER, name=name and name + "attn_conv")(inputs)
        attn = functional.reshape(attn, [-1, 1, height * width])  # [batch, height, width, 1] or [batch, 1, height, width] -> [batch, 1, height * width]
        attn = functional.softmax(attn, axis=-1)
        context = inputs if is_channels_last else functional.transpose(inputs, [0, 2, 3, 1])
        context = functional.reshape(context, [-1, height * width, filters])
        context = attn @ context  # [batch, 1, filters]
        context = functional.reshape(context, [-1, 1, 1, filters]) if is_channels_last else functional.reshape(context, [-1, filters, 1, 1])
    else:
        context = functional.reduce_mean(inputs, [height_axis, width_axis], keepdims=True)

    mlp = layers.Conv2D(reduction, kernel_size=1, use_bias=use_bias, name=name and name + "mlp_1_conv")(context)
    mlp = layers.LayerNormalization(epsilon=LAYER_NORM_EPSILON, name=name and name + "ln")(mlp)
    mlp = activation_by_name(mlp, activation=hidden_activation, name=name)
    mlp = layers.Conv2D(filters, kernel_size=1, use_bias=use_bias, name=name and name + "mlp_2_conv")(mlp)
    mlp = activation_by_name(mlp, activation=output_activation, name=name)
    return layers.Multiply(name=name and name + "out")([inputs, mlp])


def se_module(inputs, se_ratio=0.25, divisor=8, limit_round_down=0.9, activation="relu", use_bias=True, use_conv=True, name=None):
    """Squeeze-and-Excitation block, arxiv: https://arxiv.org/pdf/1709.01507.pdf"""
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    h_axis, w_axis = [1, 2] if image_data_format() == "channels_last" else [2, 3]

    # activation could be ("relu", "hard_sigmoid") for mobilenetv3
    hidden_activation, output_activation = activation if isinstance(activation, (list, tuple)) else (activation, "sigmoid")
    filters = inputs.shape[channel_axis]
    reduction = make_divisible(filters * se_ratio, divisor, limit_round_down=limit_round_down)
    # print(f"{filters = }, {se_ratio = }, {divisor = }, {reduction = }")
    se = functional.reduce_mean(inputs, [h_axis, w_axis], keepdims=True if use_conv else False)
    if use_conv:
        se = layers.Conv2D(reduction, kernel_size=1, use_bias=use_bias, kernel_initializer=CONV_KERNEL_INITIALIZER, name=name and name + "1_conv")(se)
    else:
        se = layers.Dense(reduction, use_bias=use_bias, kernel_initializer=CONV_KERNEL_INITIALIZER, name=name and name + "1_dense")(se)
    se = activation_by_name(se, activation=hidden_activation, name=name)
    if use_conv:
        se = layers.Conv2D(filters, kernel_size=1, use_bias=use_bias, kernel_initializer=CONV_KERNEL_INITIALIZER, name=name and name + "2_conv")(se)
    else:
        se = layers.Dense(filters, use_bias=use_bias, kernel_initializer=CONV_KERNEL_INITIALIZER, name=name and name + "2_dense")(se)
    se = activation_by_name(se, activation=output_activation, name=name)
    se = se if use_conv else functional.reshape(se, [-1, 1, 1, filters] if image_data_format() == "channels_last" else [-1, filters, 1, 1])
    return layers.Multiply(name=name and name + "out")([inputs, se])


def eca_module(inputs, gamma=2.0, beta=1.0, name=None, **kwargs):
    """Efficient Channel Attention block, arxiv: https://arxiv.org/pdf/1910.03151.pdf"""
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    h_axis, w_axis = [1, 2] if image_data_format() == "channels_last" else [2, 3]

    filters = inputs.shape[channel_axis]
    beta, gamma = float(beta), float(gamma)
    tt = int((np.log(float(filters)) / np.log(2.0) + beta) / gamma)
    kernel_size = max(tt if tt % 2 else tt + 1, 3)
    pad = kernel_size // 2

    nn = functional.reduce_mean(inputs, [h_axis, w_axis], keepdims=False)
    nn = functional.pad(nn, [[0, 0], [pad, pad]])
    nn = functional.expand_dims(nn, channel_axis)

    nn = layers.Conv1D(1, kernel_size=kernel_size, strides=1, padding="VALID", use_bias=False, name=name and name + "conv1d")(nn)
    nn = functional.squeeze(nn, axis=channel_axis)
    nn = activation_by_name(nn, activation="sigmoid", name=name)
    nn = nn[:, None, None] if image_data_format() == "channels_last" else nn[:, :, None, None]
    # print(f"{inputs.shape = }, {nn.shape = }")
    return layers.Multiply(name=name and name + "out")([inputs, nn])


def drop_connect_rates_split(num_blocks, start=0.0, end=0.0):
    """split drop connect rate in range `(start, end)` according to `num_blocks`"""
    drop_connect_rates = functional.split(functional.linspace(start, end, sum(num_blocks)), num_blocks)
    return [ii.numpy().tolist() for ii in drop_connect_rates]


def drop_block(inputs, drop_rate=0, name=None):
    """Stochastic Depth block by Dropout, arxiv: https://arxiv.org/abs/1603.09382"""
    if drop_rate > 0:
        noise_shape = [None] + [1] * (len(inputs.shape) - 1)  # [None, 1, 1, 1]
        return layers.Dropout(drop_rate, noise_shape=noise_shape, name=name and name + "drop")(inputs)
    else:
        return inputs


def addaptive_pooling_2d(inputs, output_size, reduce="mean", data_format="auto", name=None):
    """Auto set `pool_size` and `strides` for `MaxPool2D` or `AvgPool2D` fitting `output_size`.
    (in_height - (pool_size - strides)) / strides == out_height
    condition: pool_size >= strides, pool_size != 0, strides != 0
    strides being as large as possible: strides == in_height // out_height
    ==> pool_size = in_height - (out_height - 1) * strides, not in_height % strides, in case in_height == strides  will be 0
    """
    data_format = image_data_format() if data_format == "auto" else data_format
    height, width = inputs.shape[1:-1] if image_data_format() == "channels_last" else inputs.shape[2:]
    h_bins, w_bins = output_size[:2] if isinstance(output_size, (list, tuple)) else (output_size, output_size)
    reduce_function = layers.MaxPool2D if reduce.lower() == "max" else layers.AvgPool2D

    h_strides, w_strides = height // h_bins, width // w_bins
    h_pool_size, w_pool_size = height - (h_bins - 1) * h_strides, width - (w_bins - 1) * w_strides
    # print(f"{inputs.shape = }, {h_pool_size = }, {w_pool_size = }, {h_strides = }, {w_strides = }")
    return reduce_function(pool_size=(h_pool_size, w_pool_size), strides=(h_strides, w_strides), name=name and name + "pool")(inputs)


""" Other layers / functions """


@backend.register_keras_serializable(package="kecamCommon")
def __anti_alias_downsample_initializer__(weight_shape, dtype="float32"):
    import numpy as np

    kernel_size, channel = (weight_shape[0], weight_shape[2]) if backend.image_data_format() == "channels_last" else (weight_shape[2], weight_shape[0])
    ww = np.array(np.poly1d((0.5, 0.5)) ** (kernel_size - 1)).astype("float32")
    ww = np.expand_dims(ww, 0) * np.expand_dims(ww, 1)
    if backend.image_data_format() == "channels_last":
        ww = np.repeat(ww[:, :, None, None], channel, axis=-2)
    else:
        ww = np.repeat(ww[None, None, :, :], channel, axis=0)
    return functional.convert_to_tensor(ww, dtype=dtype)


def anti_alias_downsample(inputs, kernel_size=3, strides=2, padding="SAME", trainable=False, name=None):
    """DepthwiseConv2D performing anti-aliasing downsample, arxiv: https://arxiv.org/pdf/1904.11486.pdf"""
    return layers.DepthwiseConv2D(
        kernel_size=kernel_size,
        strides=strides,
        padding="SAME",
        use_bias=False,
        trainable=trainable,
        depthwise_initializer=__anti_alias_downsample_initializer__,
        name=name and name + "anti_alias_down",
    )(inputs)


def make_divisible(vv, divisor=4, min_value=None, limit_round_down=0.9):
    """Copied from https://github.com/tensorflow/models/blob/master/research/slim/nets/mobilenet/mobilenet.py"""
    if min_value is None:
        min_value = divisor
    new_v = max(min_value, int(vv + divisor / 2) // divisor * divisor)
    # Make sure that round down does not go down by more than 10%.
    if new_v < limit_round_down * vv:
        new_v += divisor
    return new_v


@backend.register_keras_serializable(package="kecamCommon")
def __unfold_filters_initializer__(weight_shape, dtype="float32"):
    kernel_size = weight_shape[0]
    kernel_out = kernel_size * kernel_size
    ww = functional.reshape(functional.eye(kernel_out), [kernel_size, kernel_size, 1, kernel_out])
    if len(weight_shape) == 5:  # Conv3D or Conv3DTranspose
        ww = functional.expand_dims(ww, 2)
    return ww


def fold_by_conv2d_transpose(patches, output_shape=None, kernel_size=3, strides=2, dilation_rate=1, padding="SAME", compressed="auto", name=None):
    paded = kernel_size // 2 if padding else 0
    if compressed == "auto":
        compressed = True if len(patches.shape) == 4 else False

    if compressed:
        _, hh, ww, cc = patches.shape
        channel = cc // kernel_size // kernel_size
        conv_rr = functional.reshape(patches, [-1, hh * ww, kernel_size * kernel_size, channel])
    else:
        _, hh, ww, _, _, channel = patches.shape
        # conv_rr = patches
        conv_rr = functional.reshape(patches, [-1, hh * ww, kernel_size * kernel_size, channel])
    conv_rr = functional.transpose(conv_rr, [0, 3, 1, 2])  # [batch, channnel, hh * ww, kernel * kernel]
    conv_rr = functional.reshape(conv_rr, [-1, hh, ww, kernel_size * kernel_size])

    convtrans_rr = layers.Conv2DTranspose(
        filters=1,
        kernel_size=kernel_size,
        strides=strides,
        dilation_rate=dilation_rate,
        padding="VALID",
        output_padding=paded,
        use_bias=False,
        trainable=False,
        kernel_initializer=__unfold_filters_initializer__,
        name=name and name + "fold_convtrans",
    )(conv_rr)

    out = functional.reshape(convtrans_rr[..., 0], [-1, channel, convtrans_rr.shape[1], convtrans_rr.shape[2]])
    out = functional.transpose(out, [0, 2, 3, 1])
    if output_shape is None:
        output_shape = [-paded, -paded]
    else:
        output_shape = [output_shape[0] + paded, output_shape[1] + paded]
    out = out[:, paded : output_shape[0], paded : output_shape[1]]
    return out


@backend.register_keras_serializable(package="kecamCommon")
class CompatibleExtractPatches(layers.Layer):
    def __init__(self, sizes=3, strides=2, rates=1, padding="SAME", compressed=True, force_conv=False, **kwargs):
        super().__init__(**kwargs)
        self.sizes, self.strides, self.rates, self.padding = sizes, strides, rates, padding
        self.compressed, self.force_conv = compressed, force_conv

        self.kernel_size = sizes[1] if isinstance(sizes, (list, tuple)) else sizes
        self.strides = strides[1] if isinstance(strides, (list, tuple)) else strides
        self.dilation_rate = rates[1] if isinstance(rates, (list, tuple)) else rates
        self.filters = self.kernel_size * self.kernel_size

        if backend.backend() == "tensorflow":
            import tensorflow as tf

            if len(tf.config.experimental.list_logical_devices("TPU")) != 0 or self.force_conv:
                self.use_conv = True
            else:
                self.use_conv = False
        else:
            self.use_conv = force_conv

        self.use_layer_as_module = True  # For PyTorch backend

    def build(self, input_shape):
        _, self.height, self.width, self.channel = input_shape
        if self.padding.upper() == "SAME":
            pad = self.kernel_size // 2
            self.pad_value = [[0, 0], [pad, pad], [pad, pad], [0, 0]]
            self.height, self.width = self.height + pad * 2, self.width + pad * 2

        if self.use_conv:
            self.conv = layers.Conv2D(
                filters=self.filters,
                kernel_size=self.kernel_size,
                strides=self.strides,
                dilation_rate=self.dilation_rate,
                padding="VALID",
                use_bias=False,
                trainable=False,
                kernel_initializer=__unfold_filters_initializer__,
                name=self.name and self.name + "unfold_conv",
            )
            self.conv.build([None, *input_shape[1:-1], 1])
        else:
            self._sizes_ = [1, self.kernel_size, self.kernel_size, 1]
            self._strides_ = [1, self.strides, self.strides, 1]
            self._rates_ = [1, self.dilation_rate, self.dilation_rate, 1]
        # output_size = backend.compute_conv_output_size([self.height, self.width], self.kernel_size, self.strides, self.padding, self.dilation_rate)
        # self.output_height, self.output_width = output_size
        super().build(input_shape)

    def call(self, inputs):
        if self.padding.upper() == "SAME":
            inputs = functional.pad(inputs, self.pad_value)

        if self.use_conv:
            merge_channel = functional.transpose(inputs, [0, 3, 1, 2])
            merge_channel = functional.reshape(merge_channel, [-1, self.height, self.width, 1])
            conv_rr = self.conv(merge_channel)

            # TFLite not supporting `tf.transpose` with len(perm) > 4...
            out = functional.reshape(conv_rr, [-1, self.channel, conv_rr.shape[1] * conv_rr.shape[2], self.filters])
            out = functional.transpose(out, [0, 2, 3, 1])  # [batch, hh * ww, kernel * kernel, channnel]
            if self.compressed:
                out = functional.reshape(out, [-1, conv_rr.shape[1], conv_rr.shape[2], self.filters * self.channel])
            else:
                out = functional.reshape(out, [-1, conv_rr.shape[1], conv_rr.shape[2], self.kernel_size, self.kernel_size, self.channel])
        else:
            out = functional.extract_patches(inputs, self._sizes_, self._strides_, self._rates_, "VALID")
            if not self.compressed:
                # [batch, hh, ww, kernel, kernel, channnel]
                out = functional.reshape(out, [-1, out.shape[1], out.shape[2], self.kernel_size, self.kernel_size, self.channel])
        return out

    def get_config(self):
        base_config = super().get_config()
        base_config.update(
            {
                "sizes": self.sizes,
                "strides": self.strides,
                "rates": self.rates,
                "padding": self.padding,
                "compressed": self.compressed,
                "force_conv": self.force_conv,
            }
        )
        return base_config


class PreprocessInput:
    """`rescale_mode` `torch` means `(image - [0.485, 0.456, 0.406]) / [0.229, 0.224, 0.225]`, `tf` means `(image - 0.5) / 0.5`"""

    def __init__(self, input_shape=(224, 224, 3), rescale_mode="torch"):
        self.set_input_shape(input_shape)
        self.set_rescale_mode(rescale_mode)

    def set_input_shape(self, input_shape):
        input_shape = input_shape[1:] if len(input_shape) == 4 else input_shape
        if None in input_shape:
            self.input_shape = (None, None)  # Dynamic input_shape
        else:
            channel_axis, channel_dim = min(enumerate(input_shape), key=lambda xx: xx[1])  # Assume the smallest value is the channel dimension
            self.input_shape = [dim for axis, dim in enumerate(input_shape) if axis != channel_axis]

    def set_rescale_mode(self, rescale_mode):
        if isinstance(rescale_mode, (list, tuple)):  # Specific mean and std
            mean, std = rescale_mode
        elif rescale_mode == "torch":
            mean = np.array([0.485, 0.456, 0.406]).astype("float32") * 255.0
            std = np.array([0.229, 0.224, 0.225]).astype("float32") * 255.0
            if backend.image_data_format() != "channels_last":
                mean, std = mean[:, None, None], std[:, None, None]
        elif rescale_mode == "tf":  # [0, 255] -> [-1, 1]
            mean, std = 127.5, 127.5
            # mean, std = 127.5, 128.0
        elif rescale_mode == "tf128":  # [0, 255] -> [-1, 1]
            mean, std = 128.0, 128.0
        elif rescale_mode == "raw01":
            mean, std = 0, 255.0  # [0, 255] -> [0, 1]
        else:
            mean, std = 0, 1  # raw inputs [0, 255]
        self.mean, self.std, self.rescale_mode = mean, std, rescale_mode

    def __call__(self, image, resize_method="bilinear", resize_antialias=False, input_shape=None):
        if input_shape is not None:
            self.set_input_shape(input_shape)
        images = np.array([image] if len(np.shape(image)) == 3 else image).astype("float32")
        images = (images * 255) if images.max() < 2 else images

        images = images if backend.image_data_format() == "channels_last" else images.transpose([0, 3, 1, 2])
        images = functional.convert_to_tensor(images)
        images = functional.resize(images, self.input_shape, method=resize_method, antialias=resize_antialias)
        images = (images - self.mean) / self.std
        return images


def add_pre_post_process(model, rescale_mode="tf", input_shape=None, post_process=None):
    from keras_cv_attention_models.imagenet.eval_func import decode_predictions

    input_shape = model.input_shape[1:] if input_shape is None else input_shape
    model.preprocess_input = PreprocessInput(input_shape, rescale_mode=rescale_mode)
    model.decode_predictions = decode_predictions if post_process is None else post_process
    model.rescale_mode = rescale_mode



# Copyright 2021 Google Research. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""EfficientNet V1 and V2 model.

[1] Mingxing Tan, Quoc V. Le
  EfficientNet: Rethinking Model Scaling for Convolutional Neural Networks.
  ICML'19, https://arxiv.org/abs/1905.11946

[2] Mingxing Tan, Quoc V. Le
  EfficientNetV2: Smaller Models and Faster Training.
  https://arxiv.org/abs/2104.00298
"""
import sys

sys.path.append("../automl/efficientnetv2")

import copy
import itertools
import math
import os

from absl import logging
import numpy as np
import tensorflow as tf

import effnetv2_configs
import hparams
import utils


def conv_kernel_initializer(shape, dtype=None, partition_info=None):
    """Initialization for convolutional kernels.

    The main difference with tf.variance_scaling_initializer is that
    tf.variance_scaling_initializer uses a truncated normal with an uncorrected
    standard deviation, whereas here we use a normal distribution. Similarly,
    tf.initializers.variance_scaling uses a truncated normal with
    a corrected standard deviation.

    Args:
      shape: shape of variable
      dtype: dtype of variable
      partition_info: unused

    Returns:
      an initialization for the variable
    """
    del partition_info
    kernel_height, kernel_width, _, out_filters = shape
    fan_out = int(kernel_height * kernel_width * out_filters)
    return tf.random.normal(shape, mean=0.0, stddev=np.sqrt(2.0 / fan_out), dtype=dtype)


def dense_kernel_initializer(shape, dtype=None, partition_info=None):
    """Initialization for dense kernels.

    This initialization is equal to
      tf.variance_scaling_initializer(scale=1.0/3.0, mode='fan_out',
                                      distribution='uniform').
    It is written out explicitly here for clarity.

    Args:
      shape: shape of variable
      dtype: dtype of variable
      partition_info: unused

    Returns:
      an initialization for the variable
    """
    del partition_info
    init_range = 1.0 / np.sqrt(shape[1])
    return tf.random.uniform(shape, -init_range, init_range, dtype=dtype)


def round_filters(filters, mconfig, skip=False):
    """Round number of filters based on depth multiplier."""
    multiplier = mconfig.width_coefficient
    divisor = mconfig.depth_divisor
    min_depth = mconfig.min_depth
    if skip or not multiplier:
        return filters

    filters *= multiplier
    min_depth = min_depth or divisor
    new_filters = max(min_depth, int(filters + divisor / 2) // divisor * divisor)
    return int(new_filters)


def round_repeats(repeats, multiplier, skip=False):
    """Round number of filters based on depth multiplier."""
    if skip or not multiplier:
        return repeats
    return int(math.ceil(multiplier * repeats))


class SE(tf.keras.layers.Layer):
    """Squeeze-and-excitation layer."""

    def __init__(self, mconfig, se_filters, output_filters, name=None):
        super().__init__(name=name)

        self._local_pooling = mconfig.local_pooling
        self._data_format = mconfig.data_format
        self._act = utils.get_act_fn(mconfig.act_fn)

        # Squeeze and Excitation layer.
        self._se_reduce = tf.keras.layers.Conv2D(
            se_filters, kernel_size=1, strides=1, kernel_initializer=conv_kernel_initializer, padding="same", data_format=self._data_format, use_bias=True
        )
        self._se_expand = tf.keras.layers.Conv2D(
            output_filters, kernel_size=1, strides=1, kernel_initializer=conv_kernel_initializer, padding="same", data_format=self._data_format, use_bias=True
        )

    def call(self, inputs):
        h_axis, w_axis = [2, 3] if self._data_format == "channels_first" else [1, 2]
        if self._local_pooling:
            se_tensor = tf.nn.avg_pool(inputs, ksize=[1, inputs.shape[h_axis], inputs.shape[w_axis], 1], strides=[1, 1, 1, 1], padding="VALID")
        else:
            se_tensor = tf.reduce_mean(inputs, [h_axis, w_axis], keepdims=True)
        se_tensor = self._se_expand(self._act(self._se_reduce(se_tensor)))
        logging.info("Built SE %s : %s", self.name, se_tensor.shape)
        return tf.sigmoid(se_tensor) * inputs


class MBConvBlock(tf.keras.layers.Layer):
    """A class of MBConv: Mobile Inverted Residual Bottleneck.

    Attributes:
      endpoints: dict. A list of internal tensors.
    """

    def __init__(self, block_args, mconfig, name=None):
        """Initializes a MBConv block.

        Args:
          block_args: BlockArgs, arguments to create a Block.
          mconfig: GlobalParams, a set of global parameters.
          name: layer name.
        """
        super().__init__(name=name)

        self._block_args = copy.deepcopy(block_args)
        self._mconfig = copy.deepcopy(mconfig)
        self._local_pooling = mconfig.local_pooling
        self._data_format = mconfig.data_format
        self._channel_axis = 1 if self._data_format == "channels_first" else -1

        self._act = utils.get_act_fn(mconfig.act_fn)
        self._has_se = self._block_args.se_ratio is not None and 0 < self._block_args.se_ratio <= 1

        self.endpoints = None

        # Builds the block accordings to arguments.
        self._build()

    @property
    def block_args(self):
        return self._block_args

    def _build(self):
        """Builds block according to the arguments."""
        # pylint: disable=g-long-lambda
        bid = itertools.count(0)
        get_norm_name = lambda: self.name + "tpu_batch_normalization" + ("_0" if not next(bid) else "_" + str(next(bid) // 2))
        cid = itertools.count(0)
        get_conv_name = lambda: self.name + "conv2d" + ("0" if not next(cid) else "_" + str(next(cid) // 2))
        # pylint: enable=g-long-lambda

        mconfig = self._mconfig
        filters = self._block_args.input_filters * self._block_args.expand_ratio
        kernel_size = self._block_args.kernel_size

        # Expansion phase. Called if not using fused convolutions and expansion
        # phase is necessary.
        if self._block_args.expand_ratio != 1:
            self._expand_conv = tf.keras.layers.Conv2D(
                filters=filters,
                kernel_size=1,
                strides=1,
                kernel_initializer=conv_kernel_initializer,
                padding="same",
                data_format=self._data_format,
                use_bias=False,
                name=get_conv_name(),
            )
            self._norm0 = utils.normalization(
                mconfig.bn_type,
                axis=self._channel_axis,
                momentum=mconfig.bn_momentum,
                epsilon=mconfig.bn_epsilon,
                groups=mconfig.gn_groups,
                name=get_norm_name(),
            )

        # Depth-wise convolution phase. Called if not using fused convolutions.
        self._depthwise_conv = tf.keras.layers.DepthwiseConv2D(
            kernel_size=kernel_size,
            strides=self._block_args.strides,
            depthwise_initializer=conv_kernel_initializer,
            padding="same",
            data_format=self._data_format,
            use_bias=False,
        )

        self._norm1 = utils.normalization(
            mconfig.bn_type, axis=self._channel_axis, momentum=mconfig.bn_momentum, epsilon=mconfig.bn_epsilon, groups=mconfig.gn_groups, name=get_norm_name()
        )

        if self._has_se:
            num_reduced_filters = max(1, int(self._block_args.input_filters * self._block_args.se_ratio))
            self._se = SE(self._mconfig, num_reduced_filters, filters)
        else:
            self._se = None

        # Output phase.
        filters = self._block_args.output_filters
        self._project_conv = tf.keras.layers.Conv2D(
            filters=filters,
            kernel_size=1,
            strides=1,
            kernel_initializer=conv_kernel_initializer,
            padding="same",
            data_format=self._data_format,
            use_bias=False,
            name=get_conv_name(),
        )
        self._norm2 = utils.normalization(
            mconfig.bn_type, axis=self._channel_axis, momentum=mconfig.bn_momentum, epsilon=mconfig.bn_epsilon, groups=mconfig.gn_groups, name=get_norm_name()
        )

    def residual(self, inputs, x, training, survival_prob):
        if self._block_args.strides == 1 and self._block_args.input_filters == self._block_args.output_filters:
            x = tf.keras.layers.Add()([inputs, x])

        return x

    def call(self, inputs, training, survival_prob=None):
        """Implementation of call().

        Args:
          inputs: the inputs tensor.
          training: boolean, whether the model is constructed for training.
          survival_prob: float, between 0 to 1, drop connect rate.

        Returns:
          A output tensor.
        """
        logging.info("Block %s input shape: %s (%s)", self.name, inputs.shape, inputs.dtype)
        x = inputs
        if self._block_args.expand_ratio != 1:
            x = self._act(self._norm0(self._expand_conv(x), training=training))
            logging.info("Expand shape: %s", x.shape)

        x = self._act(self._norm1(self._depthwise_conv(x), training=training))
        logging.info("DWConv shape: %s", x.shape)

        if self._mconfig.conv_dropout and self._block_args.expand_ratio > 1:
            x = tf.keras.layers.Dropout(self._mconfig.conv_dropout)(x, training=training)

        if self._se:
            x = self._se.call(x)

        self.endpoints = {"expansion_output": x}

        x = self._norm2(self._project_conv(x), training=training)
        x = self.residual(inputs, x, training, survival_prob)

        logging.info("Project shape: %s", x.shape)
        return x


class FusedMBConvBlock(MBConvBlock):
    """Fusing the proj conv1x1 and depthwise_conv into a conv2d."""

    def _build(self):
        """Builds block according to the arguments."""
        # pylint: disable=g-long-lambda
        bid = itertools.count(0)
        get_norm_name = lambda: self.name + "tpu_batch_normalization" + ("_0" if not next(bid) else "_" + str(next(bid) // 2))
        cid = itertools.count(0)
        get_conv_name = lambda: self.name + "conv2d" + ("0" if not next(cid) else "_" + str(next(cid) // 2))
        # pylint: enable=g-long-lambda

        mconfig = self._mconfig
        block_args = self._block_args
        filters = block_args.input_filters * block_args.expand_ratio
        kernel_size = block_args.kernel_size
        if block_args.expand_ratio != 1:
            # Expansion phase:
            self._expand_conv = tf.keras.layers.Conv2D(
                filters,
                kernel_size=kernel_size,
                strides=block_args.strides,
                kernel_initializer=conv_kernel_initializer,
                padding="same",
                use_bias=False,
                name=get_conv_name(),
            )
            self._norm0 = utils.normalization(
                mconfig.bn_type,
                axis=self._channel_axis,
                momentum=mconfig.bn_momentum,
                epsilon=mconfig.bn_epsilon,
                groups=mconfig.gn_groups,
                name=get_norm_name(),
            )

        if self._has_se:
            num_reduced_filters = max(1, int(block_args.input_filters * block_args.se_ratio))
            self._se = SE(mconfig, num_reduced_filters, filters)
        else:
            self._se = None
        # Output phase:
        filters = block_args.output_filters
        self._project_conv = tf.keras.layers.Conv2D(
            filters,
            kernel_size=1 if block_args.expand_ratio != 1 else kernel_size,
            strides=1 if block_args.expand_ratio != 1 else block_args.strides,
            kernel_initializer=conv_kernel_initializer,
            padding="same",
            use_bias=False,
            name=get_conv_name(),
        )
        self._norm1 = utils.normalization(
            mconfig.bn_type, axis=self._channel_axis, momentum=mconfig.bn_momentum, epsilon=mconfig.bn_epsilon, groups=mconfig.gn_groups, name=get_norm_name()
        )

    def call(self, inputs, training, survival_prob=None):
        """Implementation of call().

        Args:
          inputs: the inputs tensor.
          training: boolean, whether the model is constructed for training.
          survival_prob: float, between 0 to 1, drop connect rate.

        Returns:
          A output tensor.
        """
        logging.info("Block %s  input shape: %s", self.name, inputs.shape)
        x = inputs
        if self._block_args.expand_ratio != 1:
            x = self._act(self._norm0(self._expand_conv(x), training=training))
        logging.info("Expand shape: %s", x.shape)

        self.endpoints = {"expansion_output": x}

        if self._mconfig.conv_dropout and self._block_args.expand_ratio > 1:
            x = tf.keras.layers.Dropout(self._mconfig.conv_dropout)(x, training)

        if self._se:
            x = self._se.call(x)

        x = self._norm1(self._project_conv(x), training=training)
        if self._block_args.expand_ratio == 1:
            x = self._act(x)  # add act if no expansion.

        x = self.residual(inputs, x, training, survival_prob)
        logging.info("Project shape: %s", x.shape)
        return x


class Stem(tf.keras.layers.Layer):
    """Stem layer at the begining of the network."""

    def __init__(self, mconfig, stem_filters, name=None):
        super().__init__(name=name)
        self._conv_stem = tf.keras.layers.Conv2D(
            filters=round_filters(stem_filters, mconfig),
            kernel_size=3,
            strides=2,
            kernel_initializer=conv_kernel_initializer,
            padding="same",
            data_format=mconfig.data_format,
            use_bias=False,
        )
        self._norm = utils.normalization(
            mconfig.bn_type,
            axis=(1 if mconfig.data_format == "channels_first" else -1),
            momentum=mconfig.bn_momentum,
            epsilon=mconfig.bn_epsilon,
            groups=mconfig.gn_groups,
            name="stem_norm",
        )
        self._act = utils.get_act_fn(mconfig.act_fn)

    def call(self, inputs, training):
        return self._act(self._norm(self._conv_stem(inputs), training=training))


class Head(tf.keras.layers.Layer):
    """Head layer for network outputs."""

    def __init__(self, mconfig, name=None):
        super().__init__(name=name)

        self.endpoints = {}
        self._mconfig = mconfig

        self._conv_head = tf.keras.layers.Conv2D(
            filters=round_filters(mconfig.feature_size or 1280, mconfig),
            kernel_size=1,
            strides=1,
            kernel_initializer=conv_kernel_initializer,
            padding="same",
            data_format=mconfig.data_format,
            use_bias=False,
        )
        self._norm = utils.normalization(
            mconfig.bn_type,
            axis=(1 if mconfig.data_format == "channels_first" else -1),
            momentum=mconfig.bn_momentum,
            epsilon=mconfig.bn_epsilon,
            groups=mconfig.gn_groups,
            name="head_norm",
        )
        self._act = utils.get_act_fn(mconfig.act_fn)

        self._avg_pooling = tf.keras.layers.GlobalAveragePooling2D(data_format=mconfig.data_format)

        if mconfig.dropout_rate > 0:
            self._dropout = tf.keras.layers.Dropout(mconfig.dropout_rate)
        else:
            self._dropout = None

        self.h_axis, self.w_axis = [2, 3] if mconfig.data_format == "channels_first" else [1, 2]

    def call(self, inputs, training):
        """Call the layer."""
        outputs = self._act(self._norm(self._conv_head(inputs), training=training))
        self.endpoints["head_1x1"] = outputs

        if self._mconfig.local_pooling:
            shape = outputs.get_shape().as_list()
            kernel_size = [1, shape[self.h_axis], shape[self.w_axis], 1]
            outputs = tf.nn.avg_pool(outputs, ksize=kernel_size, strides=[1, 1, 1, 1], padding="VALID")
            self.endpoints["pooled_features"] = outputs
            if self._dropout:
                outputs = self._dropout(outputs, training=training)
            self.endpoints["global_pool"] = outputs
            if self._fc:
                outputs = tf.squeeze(outputs, [self.h_axis, self.w_axis])
                outputs = self._fc(outputs)
            self.endpoints["head"] = outputs
        else:
            outputs = self._avg_pooling(outputs)
            self.endpoints["pooled_features"] = outputs
            if self._dropout:
                outputs = self._dropout(outputs, training=training)
            self.endpoints["head"] = outputs
        return outputs


class EffNetV2Model(tf.keras.Model):
    """A class implements tf.keras.Model.

    Reference: https://arxiv.org/abs/1807.11626
    """

    def __init__(self, model_name="efficientnetv2-s", model_config=None, include_top=True, num_classes=1000, name=None):
        """Initializes an `Model` instance.

        Args:
          model_name: A string of model name.
          model_config: A dict of model configurations or a string of hparams.
          include_top: If True, include the top layer for classification.
          name: A string of layer name.

        Raises:
          ValueError: when blocks_args is not specified as a list.
        """
        super().__init__(name=name or model_name)
        cfg = copy.deepcopy(hparams.base_config)
        if model_name:
            cfg.override(effnetv2_configs.get_model_config(model_name))
        cfg.model.override(model_config)
        self.cfg = cfg
        self._mconfig = cfg.model
        self.endpoints = None
        self.include_top = include_top
        self.num_classes = num_classes
        self._build()

    def _build(self):
        """Builds a model."""
        self._blocks = []

        # Stem part.
        self._stem = Stem(self._mconfig, self._mconfig.blocks_args[0].input_filters)

        # Builds blocks.
        block_id = itertools.count(0)
        block_name = lambda: "blocks_%d" % next(block_id)
        for block_args in self._mconfig.blocks_args:
            assert block_args.num_repeat > 0
            # Update block input and output filters based on depth multiplier.
            input_filters = round_filters(block_args.input_filters, self._mconfig)
            output_filters = round_filters(block_args.output_filters, self._mconfig)
            repeats = round_repeats(block_args.num_repeat, self._mconfig.depth_coefficient)
            block_args.update(dict(input_filters=input_filters, output_filters=output_filters, num_repeat=repeats))

            # The first block needs to take care of stride and filter size increase.
            conv_block = {0: MBConvBlock, 1: FusedMBConvBlock}[block_args.conv_type]
            self._blocks.append(conv_block(block_args, self._mconfig, name=block_name()))
            if block_args.num_repeat > 1:  # rest of blocks with the same block_arg
                # pylint: disable=protected-access
                block_args.input_filters = block_args.output_filters
                block_args.strides = 1
                # pylint: enable=protected-access
            for _ in range(block_args.num_repeat - 1):
                self._blocks.append(conv_block(block_args, self._mconfig, name=block_name()))

        # Head part.
        self._head = Head(self._mconfig)

        # top part for classification
        if self.include_top and self.num_classes:
            self._fc = tf.keras.layers.Dense(
                self.num_classes, kernel_initializer=dense_kernel_initializer, bias_initializer=tf.constant_initializer(self._mconfig.headbias or 0)
            )
        else:
            self._fc = None

    def summary(self, input_shape=(224, 224, 3), **kargs):
        x = tf.keras.Input(shape=input_shape)
        model = tf.keras.Model(inputs=[x], outputs=self.call(x, training=True))
        return model.summary()

    def get_model_with_inputs(self, inputs, **kargs):
        model = tf.keras.Model(inputs=[inputs], outputs=self.call(inputs, training=True))
        return model

    def call(self, inputs, training, with_endpoints=False):
        """Implementation of call().

        Args:
          inputs: input tensors.
          training: boolean, whether the model is constructed for training.
          with_endpoints: If true, return a list of endpoints.

        Returns:
          output tensors.
        """
        outputs = None
        self.endpoints = {}
        reduction_idx = 0

        # Calls Stem layers
        outputs = self._stem.call(inputs, training)
        logging.info("Built stem: %s (%s)", outputs.shape, outputs.dtype)
        self.endpoints["stem"] = outputs

        # Calls blocks.
        for idx, block in enumerate(self._blocks):
            is_reduction = False  # reduction flag for blocks after the stem layer
            if (idx == len(self._blocks) - 1) or self._blocks[idx + 1].block_args.strides > 1:
                is_reduction = True
                reduction_idx += 1

            survival_prob = self._mconfig.survival_prob
            if survival_prob:
                drop_rate = 1.0 - survival_prob
                survival_prob = 1.0 - drop_rate * float(idx) / len(self._blocks)
                logging.info("block_%s survival_prob: %s", idx, survival_prob)
            outputs = block.call(outputs, training=training, survival_prob=survival_prob)
            self.endpoints["block_%s" % idx] = outputs
            if is_reduction:
                self.endpoints["reduction_%s" % reduction_idx] = outputs
            if block.endpoints:
                for k, v in block.endpoints.items():
                    self.endpoints["block_%s/%s" % (idx, k)] = v
                    if is_reduction:
                        self.endpoints["reduction_%s/%s" % (reduction_idx, k)] = v
        self.endpoints["features"] = outputs

        # Head to obtain the final feature.
        outputs = self._head.call(outputs, training)
        self.endpoints.update(self._head.endpoints)

        # Calls final dense layers and returns logits.
        if self._fc:
            with tf.name_scope("head"):  # legacy
                outputs = self._fc(outputs)

        if with_endpoints:  # Use for building sequential models.
            return [outputs] + list(
                filter(
                    lambda endpoint: endpoint is not None,
                    [
                        self.endpoints.get("reduction_1"),
                        self.endpoints.get("reduction_2"),
                        self.endpoints.get("reduction_3"),
                        self.endpoints.get("reduction_4"),
                        self.endpoints.get("reduction_5"),
                    ],
                )
            )

        return outputs


def get_model(model_name, model_config=None, include_top=True, weights="imagenet", training=True, with_endpoints=False, **kwargs):
    """Get a EfficientNet V1 or V2 model instance.

    This is a simply utility for finetuning or inference.

    Args:
      model_name: a string such as 'efficientnetv2-s' or 'efficientnet-b0'.
      model_config: A dict of model configurations or a string of hparams.
      include_top: whether to include the final dense layer for classification.
      weights: One of None (random initialization),
        'imagenet' (pretrained on ImageNet),
        'imagenet21k' (pretrained on Imagenet21k),
        'imagenet21k-ft1k' (pretrained on 21k and finetuned on 1k),
        'jft' (trained with non-labelled JFT-300),
        or the path to the weights file to be loaded. Defaults to 'imagenet'.
      training: If true, all model variables are trainable.
      with_endpoints: whether to return all intermedia endpoints.
      **kwargs: additional parameters for keras model, such as name=xx.

    Returns:
      A single tensor if with_endpoints if False; otherwise, a list of tensor.
    """
    net = EffNetV2Model(model_name, model_config, include_top, **kwargs)
    net(tf.keras.Input(shape=(None, None, 3)), training=training, with_endpoints=with_endpoints)

    if not weights:  # pylint: disable=g-bool-id-comparison
        return net

    v2url = "https://storage.googleapis.com/cloud-tpu-checkpoints/efficientnet/v2/"
    v1url = "https://storage.googleapis.com/cloud-tpu-checkpoints/efficientnet/advprop/"
    v1jfturl = "https://storage.googleapis.com/cloud-tpu-checkpoints/efficientnet/noisystudent/"
    pretrained_ckpts = {
        # EfficientNet V2.
        "efficientnetv2-s": {
            "imagenet": v2url + "efficientnetv2-s.tgz",
            "imagenet21k": v2url + "efficientnetv2-s-21k.tgz",
            "imagenet21k-ft1k": v2url + "efficientnetv2-s-21k-ft1k.tgz",
        },
        "efficientnetv2-m": {
            "imagenet": v2url + "efficientnetv2-m.tgz",
            "imagenet21k": v2url + "efficientnetv2-m-21k.tgz",
            "imagenet21k-ft1k": v2url + "efficientnetv2-m-21k-ft1k.tgz",
        },
        "efficientnetv2-l": {
            "imagenet": v2url + "efficientnetv2-l.tgz",
            "imagenet21k": v2url + "efficientnetv2-l-21k.tgz",
            "imagenet21k-ft1k": v2url + "efficientnetv2-l-21k-ft1k.tgz",
        },
        "efficientnetv2-xl": {
            # no imagenet ckpt.
            "imagenet21k": v2url + "efficientnetv2-xl-21k.tgz",
            "imagenet21k-ft1k": v2url + "efficientnetv2-xl-21k-ft1k.tgz",
        },
        "efficientnetv2-b0": {
            "imagenet": v2url + "efficientnetv2-b0.tgz",
            "imagenet21k": v2url + "efficientnetv2-b0-21k.tgz",
            "imagenet21k-ft1k": v2url + "efficientnetv2-b0-21k-ft1k.tgz",
        },
        "efficientnetv2-b1": {
            "imagenet": v2url + "efficientnetv2-b1.tgz",
            "imagenet21k": v2url + "efficientnetv2-b1-21k.tgz",
            "imagenet21k-ft1k": v2url + "efficientnetv2-b1-21k-ft1k.tgz",
        },
        "efficientnetv2-b2": {
            "imagenet": v2url + "efficientnetv2-b2.tgz",
            "imagenet21k": v2url + "efficientnetv2-b2-21k.tgz",
            "imagenet21k-ft1k": v2url + "efficientnetv2-b2-21k-ft1k.tgz",
        },
        "efficientnetv2-b3": {
            "imagenet": v2url + "efficientnetv2-b3.tgz",
            "imagenet21k": v2url + "efficientnetv2-b3-21k.tgz",
            "imagenet21k-ft1k": v2url + "efficientnetv2-b3-21k-ft1k.tgz",
        },
        # EfficientNet V1.
        "efficientnet-b0": {
            "imagenet": v1url + "efficientnet-b0.tar.gz",
            "jft": v1jfturl + "noisy_student_efficientnet-b0.tar.gz",
        },
        "efficientnet-b1": {
            "imagenet": v1url + "efficientnet-b1.tar.gz",
            "jft": v1jfturl + "noisy_student_efficientnet-b1.tar.gz",
        },
        "efficientnet-b2": {
            "imagenet": v1url + "efficientnet-b2.tar.gz",
            "jft": v1jfturl + "noisy_student_efficientnet-b2.tar.gz",
        },
        "efficientnet-b3": {
            "imagenet": v1url + "efficientnet-b3.tar.gz",
            "jft": v1jfturl + "noisy_student_efficientnet-b3.tar.gz",
        },
        "efficientnet-b4": {
            "imagenet": v1url + "efficientnet-b4.tar.gz",
            "jft": v1jfturl + "noisy_student_efficientnet-b4.tar.gz",
        },
        "efficientnet-b5": {
            "imagenet": v1url + "efficientnet-b5.tar.gz",
            "jft": v1jfturl + "noisy_student_efficientnet-b5.tar.gz",
        },
        "efficientnet-b6": {
            "imagenet": v1url + "efficientnet-b6.tar.gz",
            "jft": v1jfturl + "noisy_student_efficientnet-b6.tar.gz",
        },
        "efficientnet-b7": {
            "imagenet": v1url + "efficientnet-b7.tar.gz",
            "jft": v1jfturl + "noisy_student_efficientnet-b7.tar.gz",
        },
        "efficientnet-b8": {
            "imagenet": v1url + "efficientnet-b8.tar.gz",
        },
        "efficientnet-l2": {
            "jft": v1jfturl + "noisy_student_efficientnet-l2_475.tar.gz",
        },
    }

    if model_name in pretrained_ckpts and weights in pretrained_ckpts[model_name]:
        url = pretrained_ckpts[model_name][weights]
        fname = os.path.basename(url).split(".")[0]
        pretrained_ckpt = tf.keras.utils.get_file(fname, url, untar=True)
    else:
        pretrained_ckpt = weights

    if tf.io.gfile.isdir(pretrained_ckpt):
        pretrained_ckpt = tf.train.latest_checkpoint(pretrained_ckpt)
    net.load_weights(pretrained_ckpt)
    return net


if __name__ == "__main__":
    import os
    import sys
    import tensorflow as tf
    import numpy as np
    from tensorflow import keras
    import argparse

    default_save_dir = "./"
    all_model_types = ["b0", "b1", "b2", "b3", "s", "m", "l", "xl"]
    all_datasets = ["imagenet", "imagenet21k", "imagenetft"]
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("-m", "--model_type", type=str, default="s", help="all or value in {}".format(all_model_types))
    parser.add_argument("-d", "--dataset", type=str, default="imagenet", help="all or value in {}".format(all_datasets))
    parser.add_argument("-s", "--save_dir", type=str, default=default_save_dir, help="Model save dir")
    parser.add_argument("-T", "--dont_save_no_top", action="store_true", help="Dont save no_top model")
    args = parser.parse_known_args(sys.argv[1:])[0]

    import datasets as orign_datasets
    import effnetv2_model as orign_effnetv2_model

    from keras_cv_attention_models import efficientnet

    """ Parameters """
    model_type_list = [args.model_type] if args.model_type != "all" else all_model_types
    dataset_list = [args.dataset] if args.dataset != "all" else all_datasets

    for model_type in model_type_list:
        for dataset in dataset_list:
            print(">>>> model_type = {}, dataset = {}".format(model_type, dataset))
            if model_type == "xl" and dataset == "imagenet":
                print(">>>> Not included")
                continue

            assert model_type in all_model_types
            assert dataset in all_datasets

            keras.backend.clear_session()

            if dataset == "imagenet21k":
                classes, dropout, survival_prob, load_weights, save_model_suffix = 21843, 1e-6, 1.0, "imagenet21k", "-21k"
            elif dataset == "imagenetft":
                classes, dropout, survival_prob, load_weights, save_model_suffix = 1000, 0.2, 0.8, "imagenet21k-ft1k", "-21k-ft1k"
            else:  # "imagenet"
                classes, dropout, survival_prob, load_weights, save_model_suffix = 1000, 0.2, 0.8, "imagenet", "-imagenet"

            print(">>>> classes = {}, dropout = {}, load_weights = {}, save_model_suffix = {}".format(classes, dropout, load_weights, save_model_suffix))

            """ Define Keras model first just to keep the names start from `0` """
            keras_model = efficientnet.EfficientNetV2(
                model_type=model_type, drop_connect_rate=0, dropout=dropout, num_classes=classes, classifier_activation=None, pretrained=None
            )

            """ Load checkpoints using official defination """
            cc = orign_datasets.get_dataset_config(dataset)
            if cc.get("model", None):
                cc.model.num_classes = cc.data.num_classes
            else:
                cc["model"] = None
            model = orign_effnetv2_model.get_model("efficientnetv2-{}".format(model_type), model_config=cc.model, weights=load_weights)

            """ Save h5 weights if no error happens """
            model.save_weights("convert_temp_aa.h5")

            """ Reload weights with this modified version """
            mm = EffNetV2Model("efficientnetv2-{}".format(model_type), num_classes=classes)
            len(mm(tf.ones([1, 224, 224, 3]), False))
            mm.load_weights("convert_temp_aa.h5")

            """ Define a new model using `mm.call`, as mm is a subclassed model, cannot be saved as h5 """
            inputs = keras.Input([None, None, 3])
            tt = keras.models.Model(inputs, mm.call(inputs, training=False))
            tt.save("convert_temp_bb.h5")  # This is already a converted one.

            """ Reload convert_temp_bb.h5 using full keras defined model """
            keras_model.load_weights("convert_temp_bb.h5")

            """ Output verification """
            fake_input = tf.random.uniform([2, 224, 224, 3])
            orign_out = model(fake_input)
            converted_out = keras_model(fake_input)
            test_result = np.allclose(orign_out.numpy(), converted_out.numpy())
            print(">>>> Allclose:", test_result)
            # Allclose: True

            assert test_result

            """ Save model and notop version """
            if not os.path.exists(args.save_dir):
                os.makedirs(args.save_dir)
            save_path = os.path.join(args.save_dir, "efficientnetv2-{}{}.h5".format(model_type, save_model_suffix))
            save_notop_path = os.path.join(args.save_dir, "efficientnetv2-{}{}-notop.h5".format(model_type, save_model_suffix))
            print(">>>> save_path:", save_path)
            keras_model.save(save_path)
            if not args.dont_save_no_top:
                print(">>>> save_notop_path:", save_notop_path)
                keras.models.Model(keras_model.inputs[0], keras_model.layers[-4].output).save(save_notop_path)

    os.remove("convert_temp_aa.h5")
    os.remove("convert_temp_bb.h5")



from keras_cv_attention_models.hornet.hornet import (
    HorNet,
    HorNetTiny,
    HorNetTinyGF,
    HorNetSmall,
    HorNetSmallGF,
    HorNetBase,
    HorNetBaseGF,
    HorNetLarge,
    HorNetLargeGF,
    ComplexDense,
    global_local_filter,
    gnconv,
)

__head_doc__ = """
Keras implementation of [Github raoyongming/hornet](https://github.com/raoyongming/hornet).
Paper [PDF 2207.14284 HorNet: Efficient High-Order Spatial Interactions with Recursive Gated Convolutions](https://arxiv.org/pdf/2207.14284.pdf).
"""

__tail_doc__ = """  layer_scale: layer scale init value. Default `-1` means not applying, any value `>=0` will add a scale value for each block output.
      [Going deeper with Image Transformers](https://arxiv.org/pdf/2103.17239.pdf).
  input_shape: it should have exactly 3 inputs channels, like `(224, 224, 3)`.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  activation: activation used in whole model, default `gelu`.
  drop_connect_rate: is used for [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382).
      Can be value like `0.2`, indicates the drop probability linearly changes from `0 --> 0.2` for `top --> bottom` layers.
      A higher value means a higher probability will drop the deep branch.
      or `0` to disable (default).
  dropout: dropout rate if top layers is included.
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
  pretrained: one of None or "imagenet".
      Will try to download and load pre-trained model weights if not None.

Returns:
    A `keras.Model` instance.
"""

HorNet.__doc__ = __head_doc__ + """
Args:
  num_blocks: number of blocks in each stack.
  embed_dim: basic hidden dims, expand * 2 for each stack.
  mlp_ratio: expand ratio for mlp blocks hidden channel.
  gn_split: int or list of int value. Number of channels to split in `gnconv` block for each stack. Default `[2, 3, 4, 5]`
  use_global_local_filter: boolean or list of boolean value. Whether using `global_local_filter` or not for each stack.
      Default `[False, False, True, True]` for `HorNet*GF` models, `False` for others.
  scale: float or list of float value. Multiply scale for depth wise output in `gnconv` block.
  model_name: string, model name.
""" + __tail_doc__ + """
Model architectures:
  | Model         | Params | FLOPs  | Input | Top1 Acc |
  | ------------- | ------ | ------ | ----- | -------- |
  | HorNetTiny    | 22.4M  | 4.01G  | 224   | 82.8     |
  | HorNetTinyGF  | 23.0M  | 3.94G  | 224   | 83.0     |
  | HorNetSmall   | 49.5M  | 8.87G  | 224   | 83.8     |
  | HorNetSmallGF | 50.4M  | 8.77G  | 224   | 84.0     |
  | HorNetBase    | 87.3M  | 15.65G | 224   | 84.2     |
  | HorNetBaseGF  | 88.4M  | 15.51G | 224   | 84.3     |
  | HorNetLarge   | 194.5M | 34.91G | 224   | 86.8     |
  | HorNetLargeGF | 196.3M | 34.72G | 224   | 87.0     |
  | HorNetLargeGF | 201.8M | 102.0G | 384   | 87.7     |
"""

HorNetTiny.__doc__ = __head_doc__ + """
Args:
""" + __tail_doc__

HorNetTinyGF.__doc__ = HorNetTiny.__doc__
HorNetSmall.__doc__ = HorNetTiny.__doc__
HorNetSmallGF.__doc__ = HorNetTiny.__doc__
HorNetBase.__doc__ = HorNetTiny.__doc__
HorNetBaseGF.__doc__ = HorNetTiny.__doc__
HorNetLarge.__doc__ = HorNetTiny.__doc__
HorNetLargeGF.__doc__ = HorNetTiny.__doc__

ComplexDense.__doc__ = __head_doc__ + """
A Dense like layer on complex domain.
Currently limited for `channels_last` format and applying weights on `[height, width, channel]` dimension.

Examples:

>>> from keras_cv_attention_models import attention_layers
>>> aa = attention_layers.ComplexDense(name="complex_dense")
>>> print(f"{aa(tf.ones([1, 23, 24, 16])).shape = }")
# aa(tf.ones([1, 23, 24, 16])).shape = TensorShape([1, 23, 24, 16])
>>> print({ii.name: ii.shape for ii in aa.weights})
# {'complex_dense/complex_weight:0': TensorShape([2, 23, 24, 16])}
"""

global_local_filter.__doc__ = __head_doc__ + """
A layer processing half of the channels with the `global filter` and the other half with 3×3 depth-wise convolutions.
The `global filter` multiplies the frequency domain features with learnable global filters, which is equivalent to
a convolution in the spatial domain with a global kernel size and circular padding.

Args:
  inputs: input tensor.

Examples:

>>> from keras_cv_attention_models import attention_layers
>>> inputs = keras.layers.Input([14, 16, 256])
>>> nn = attention_layers.global_local_filter(inputs, name="gf_")
>>> print(f"{nn.shape = }")
# nn.shape = TensorShape([None, 14, 16, 512])

>>> mm = keras.models.Model(inputs, nn)
>>> mm.summary()
>>> print({ii.name: ii.shape for ii in mm.weights})
# {'gf_pre_ln/gamma:0': TensorShape([256]),
#  'gf_pre_ln/beta:0': TensorShape([256]),
#  'gf_complex_dense/complex_weight:0': TensorShape([2, 14, 9, 128]),
#  'gf_dw_conv/depthwise_kernel:0': TensorShape([3, 3, 128, 1]),
#  'gf_post_ln/gamma:0': TensorShape([256]),
#  'gf_post_ln/beta:0': TensorShape([256])}
"""

gnconv.__doc__ = __head_doc__ + """
gnConv: Recursive Gated Convolutions, an efficient operation to achieve long-term and high-order spatial interactions.
The gnConv is built with standard convolutions, linear projections and elementwise multiplications, but has a similar
function of input-adaptive spatial mixing to self-attention.

Args:
  inputs: input tensor.
  use_global_local_filter: boolean value whether using `DepthwiseConv2D` or `global_local_filter`.
  dw_kernel_size: kernel_size for `DepthwiseConv2D` if `use_global_local_filter=False`.
  gn_split: number of channels to split. Controls the number of recursive steps.
  scale: multiply scale for `DepthwiseConv2D` or `global_local_filter` output.

Examples:

>>> from keras_cv_attention_models import attention_layers
>>> inputs = keras.layers.Input([14, 16, 256])
>>> nn = attention_layers.gnconv(inputs, use_global_local_filter=False, name="gnconv_")
>>> print(f"{nn.shape = }")
# nn.shape = TensorShape([None, 14, 16, 512])

>>> mm = keras.models.Model(inputs, nn)
>>> mm.summary()
>>> print({ii.name: ii.shape for ii in mm.weights})
# {'gnconv_pre_conv/kernel:0': TensorShape([1, 1, 256, 512]),
#  'gnconv_pre_conv/bias:0': TensorShape([512]),
#  'gnconv_list_dw_conv/depthwise_kernel:0': TensorShape([7, 7, 448, 1]),
#  'gnconv_list_dw_conv/bias:0': TensorShape([448]),
#  'gnconv_pw1_conv/kernel:0': TensorShape([1, 1, 64, 128]),
#  'gnconv_pw1_conv/bias:0': TensorShape([128]),
#  'gnconv_pw2_conv/kernel:0': TensorShape([1, 1, 128, 256]),
#  'gnconv_pw2_conv/bias:0': TensorShape([256]),
#  'gnconv_output_conv/kernel:0': TensorShape([1, 1, 256, 256]),
#  'gnconv_output_conv/bias:0': TensorShape([256])}
"""



from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, functional, models, image_data_format
from keras_cv_attention_models.attention_layers import (
    activation_by_name,
    batchnorm_with_activation,
    conv2d_no_bias,
    add_pre_post_process,
)
from keras_cv_attention_models import model_surgery
from keras_cv_attention_models.download_and_load import reload_model_weights
from keras_cv_attention_models.coco import eval_func, anchors_func

PRETRAINED_DICT = {
    "yolov7_csp": {"coco": "71a00866fce212fa5fb11a6aeccc31e2"},
    "yolov7_d6": {"coco": "f1d59775cd5fcfc687b468c559bc5678"},
    "yolov7_e6": {"coco": "46d1671a9aa074c918c4509d29c16dd4"},
    "yolov7_e6e": {"coco": "b1316a8a431b34d39fc82d5591246713"},
    "yolov7_tiny": {"coco": "62e082bb032b7836d8494070f9dcacd6"},
    "yolov7_w6": {"coco": "ea153e7121645a55968a80dada90e67c"},
    "yolov7_x": {"coco": "0749c8c7d2554965b6559226aedbe4de"},
}


""" Yolov7Backbone """
BATCH_NORM_EPSILON = 1e-3
BATCH_NORM_MOMENTUM = 0.97


def conv_bn(inputs, output_channel, kernel_size=1, strides=1, activation="swish", name=""):
    # print(f">>>> {inputs.shape = }, {output_channel = }, {kernel_size = }, {strides = }")
    nn = conv2d_no_bias(inputs, output_channel, kernel_size, strides, padding="SAME", name=name)
    return batchnorm_with_activation(nn, activation=activation, epsilon=BATCH_NORM_EPSILON, momentum=BATCH_NORM_MOMENTUM, name=name)


def __concat_stack__(inputs, filters, concats=[-1, -3, -5, -6], depth=6, mid_ratio=1.0, out_channels=-1, activation="swish", name=""):
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    first = conv_bn(inputs, filters, kernel_size=1, strides=1, activation=activation, name=name + "1_")
    second = conv_bn(inputs, filters, kernel_size=1, strides=1, activation=activation, name=name + "2_")

    gathered = [first, second]
    mid_filters = int(mid_ratio * filters)
    for id in range(depth - 2):
        nn = conv_bn(gathered[-1], mid_filters, kernel_size=3, strides=1, activation=activation, name=name + "{}_".format(id + 3))
        gathered.append(nn)
    nn = functional.concat([gathered[ii] for ii in concats], axis=channel_axis)
    out_channels = out_channels if out_channels > 0 else nn.shape[channel_axis]
    nn = conv_bn(nn, out_channels, kernel_size=1, strides=1, activation=activation, name=name + "out_")
    return nn


def concat_stack(inputs, filters, concats=None, depth=6, mid_ratio=1.0, out_channels=-1, use_additional_stack=False, activation="swish", name=""):
    concats = concats if concats is not None else [-(ii + 1) for ii in range(depth)]  # [-1, -2, -3, -4, -5, -6] if None and depth=6
    nn = __concat_stack__(inputs, filters, concats, depth=depth, mid_ratio=mid_ratio, out_channels=out_channels, activation=activation, name=name)
    if use_additional_stack:
        cur_name = name + "another_"
        parallel = __concat_stack__(inputs, filters, concats, depth=depth, mid_ratio=mid_ratio, out_channels=out_channels, activation=activation, name=cur_name)
        nn = layers.Add()([nn, parallel])
    return nn


def csp_downsample(inputs, ratio=0.5, activation="swish", name=""):
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    input_channel = inputs.shape[channel_axis]
    hidden_ratio, out_ratio = ratio if isinstance(ratio, (list, tuple)) else (ratio, ratio)
    hidden_channel, out_channel = int(input_channel * hidden_ratio), int(input_channel * out_ratio)
    pool_branch = layers.MaxPool2D(pool_size=2, strides=2, padding="SAME", name=name + "pool")(inputs)
    if out_channel == 0:
        nn = pool_branch  # Maxpool only
    else:
        pool_branch = conv_bn(pool_branch, out_channel, kernel_size=1, strides=1, activation=activation, name=name + "pool_")
        conv_branch = conv_bn(inputs, hidden_channel, kernel_size=1, strides=1, activation=activation, name=name + "conv_1_")
        conv_branch = conv_bn(conv_branch, out_channel, kernel_size=3, strides=2, activation=activation, name=name + "conv_2_")

        nn = functional.concat([conv_branch, pool_branch], axis=channel_axis)
    return nn


# Almost same with yolor, just supporting YOLOV7_Tiny with depth=1
def res_spatial_pyramid_pooling(inputs, depth=2, expansion=0.5, pool_sizes=(5, 9, 13), activation="swish", name=""):
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    input_channels = inputs.shape[channel_axis]
    hidden_channels = int(input_channels * expansion)
    short = conv_bn(inputs, hidden_channels, kernel_size=1, activation=activation, name=name + "short_")

    deep = conv_bn(inputs, hidden_channels, kernel_size=1, activation=activation, name=name + "pre_1_")
    if depth > 1:  # depth = 1 for yolov7_tiny
        deep = conv_bn(deep, hidden_channels, kernel_size=3, activation=activation, name=name + "pre_2_")
        deep = conv_bn(deep, hidden_channels, kernel_size=1, activation=activation, name=name + "pre_3_")
    pp = [layers.MaxPool2D(pool_size=ii, strides=1, padding="SAME")(deep) for ii in pool_sizes]
    deep = functional.concat([deep, *pp], axis=channel_axis)  # yolov7 SPPCSPC concat, different from yolor
    for id in range(depth - 1):  # First one is `pre`
        deep = conv_bn(deep, hidden_channels, kernel_size=1, activation=activation, name=name + "post_{}_".format(id * 2 + 1))
        deep = conv_bn(deep, hidden_channels, kernel_size=3, activation=activation, name=name + "post_{}_".format(id * 2 + 2))

    if depth == 1:  # For yolov7_tiny
        deep = conv_bn(deep, hidden_channels, kernel_size=1, activation=activation, name=name + "post_1_")

    out = functional.concat([deep, short], axis=channel_axis)
    out = conv_bn(out, hidden_channels, kernel_size=1, activation=activation, name=name + "output_")
    return out


# Same with yolor
def focus_stem(inputs, filters, kernel_size=3, strides=1, padding="valid", activation="swish", name=""):
    is_channels_last = image_data_format() == "channels_last"
    channel_axis = -1 if is_channels_last else 1
    if padding.lower() == "same":  # Handling odd input_shape
        inputs = functional.pad(inputs, [[0, 0], [0, 1], [0, 1], [0, 0]] if is_channels_last else [[0, 0], [0, 0], [0, 1], [0, 1]])
        patch_top_left = inputs[:, :-1:2, :-1:2] if is_channels_last else inputs[:, :, :-1:2, :-1:2]
        patch_top_right = inputs[:, :-1:2, 1::2] if is_channels_last else inputs[:, :, :-1:2, 1::2]
        patch_bottom_left = inputs[:, 1::2, :-1:2] if image_data_format() == "channels_last" else inputs[:, :, 1::2, :-1:2]
        patch_bottom_right = inputs[:, 1::2, 1::2] if image_data_format() == "channels_last" else inputs[:, :, 1::2, 1::2]
    else:
        patch_top_left = inputs[:, ::2, ::2] if is_channels_last else inputs[:, :, ::2, ::2]
        patch_top_right = inputs[:, ::2, 1::2] if is_channels_last else inputs[:, :, ::2, 1::2]
        patch_bottom_left = inputs[:, 1::2, ::2] if is_channels_last else inputs[:, :, 1::2, ::2]
        patch_bottom_right = inputs[:, 1::2, 1::2] if is_channels_last else inputs[:, :, 1::2, 1::2]
    nn = functional.concat([patch_top_left, patch_bottom_left, patch_top_right, patch_bottom_right], axis=channel_axis)
    nn = conv_bn(nn, filters, kernel_size=kernel_size, strides=strides, activation=activation, name=name)
    return nn


def YOLOV7Backbone(
    channels=[64, 128, 256, 256],
    stack_concats=[-1, -3, -5, -6],
    stack_depth=6,
    stack_out_ratio=1.0,
    use_additional_stack=False,
    stem_width=-1,  # -1 means using channels[0]
    stem_type="conv3",  # One of ["conv3", "focus", "conv1"], "focus" for YOLOV7_*6 models, "conv1" for YOLOV7_Tiny
    csp_downsample_ratios=[0, 0.5, 0.5, 0.5],
    out_features=[-3, -2, -1],
    spp_depth=2,
    input_shape=(512, 512, 3),
    activation="swish",
    model_name="yolov7_backbone",
):
    inputs = layers.Input(backend.align_input_shape_by_image_data_format(input_shape))
    channel_axis = -1 if image_data_format() == "channels_last" else 1

    """ Stem """
    stem_width = stem_width if stem_width > 0 else channels[0]
    if stem_type == "focus":
        nn = focus_stem(inputs, stem_width, activation=activation, name="stem_")
    elif stem_type == "conv1":
        nn = conv_bn(inputs, stem_width, kernel_size=3, strides=2, activation=activation, name="stem_")
    else:
        nn = conv_bn(inputs, stem_width // 2, kernel_size=3, strides=1, activation=activation, name="stem_1_")
        nn = conv_bn(nn, stem_width, kernel_size=3, strides=2, activation=activation, name="stem_2_")
        nn = conv_bn(nn, stem_width, kernel_size=3, strides=1, activation=activation, name="stem_3_")

    common_kwargs = {
        "concats": stack_concats,
        "depth": stack_depth,
        "mid_ratio": 1.0,
        "use_additional_stack": use_additional_stack,
        "activation": activation,
    }

    """ blocks """
    features = [nn]
    for id, (channel, csp_downsample_ratio) in enumerate(zip(channels, csp_downsample_ratios)):
        stack_name = "stack{}_".format(id + 1)
        if isinstance(csp_downsample_ratio, (list, tuple)) or 0 < csp_downsample_ratio <= 1:
            nn = csp_downsample(nn, ratio=csp_downsample_ratio, activation=activation, name=stack_name + "downsample_")
        else:
            # nn = conv_bn(nn, nn.shape[channel_axis] * 2, kernel_size=3, strides=2, activation=activation, name=stack_name + "downsample_")
            ds_channels = nn.shape[channel_axis] * 2 if csp_downsample_ratio <= 0 else csp_downsample_ratio
            nn = conv_bn(nn, ds_channels, kernel_size=3, strides=2, activation=activation, name=stack_name + "downsample_")
        out_channels = -1 if stack_out_ratio == 1 else int(channel * len(stack_concats) * stack_out_ratio)
        nn = concat_stack(nn, channel, **common_kwargs, out_channels=out_channels, name=stack_name)

        if id == len(channels) - 1:
            # add SPPCSPC block if it's the last stack
            nn = res_spatial_pyramid_pooling(nn, depth=spp_depth, activation=activation, name=stack_name + "spp_")
        features.append(nn)

    nn = [features[ii] for ii in out_features]
    model = models.Model(inputs, nn, name=model_name)
    return model


""" path aggregation fpn, using `concat_stack` instead of `csp_stack` from yolor """


def upsample_merge(inputs, hidden_channels, mid_ratio=0.5, concats=None, depth=6, use_additional_stack=False, activation="swish", name=""):
    # print(f">>>> upsample_merge inputs: {[ii.shape for ii in inputs] = }")
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    upsample = conv_bn(inputs[-1], inputs[0].shape[channel_axis], activation=activation, name=name + "up_")

    # inputs[0] = layers.UpSampling2D(size=(2, 2), interpolation="nearest", name=name + "up")(fpn_out)
    size = functional.shape(inputs[0])[1:-1] if image_data_format() == "channels_last" else functional.shape(inputs[0])[2:]
    inputs[-1] = functional.resize(upsample, size, method="nearest")
    nn = functional.concat(inputs, axis=channel_axis)
    out_channels = nn.shape[channel_axis] // 2
    hidden_channels = hidden_channels if hidden_channels > 0 else nn.shape[channel_axis] // 2
    nn = concat_stack(nn, hidden_channels, concats, depth, mid_ratio, out_channels, use_additional_stack, activation=activation, name=name)
    return nn


def downsample_merge(
    inputs, hidden_channels, mid_ratio=0.5, concats=None, depth=6, csp_downsample_ratio=1, use_additional_stack=False, activation="swish", name=""
):
    # print(f">>>> downsample_merge inputs: {[ii.shape for ii in inputs] = }")
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    if isinstance(csp_downsample_ratio, (list, tuple)) or csp_downsample_ratio > 0:
        inputs[0] = csp_downsample(inputs[0], ratio=csp_downsample_ratio, activation=activation, name=name)
    else:
        inputs[0] = conv_bn(inputs[0], inputs[-1].shape[channel_axis], kernel_size=3, strides=2, activation=activation, name=name)

    nn = functional.concat(inputs, axis=channel_axis)
    out_channels = nn.shape[channel_axis] // 2
    hidden_channels = hidden_channels if hidden_channels > 0 else nn.shape[channel_axis] // 2
    nn = concat_stack(nn, hidden_channels, concats, depth, mid_ratio, out_channels, use_additional_stack, activation=activation, name=name)
    return nn


def path_aggregation_fpn(
    features, hidden, mid_ratio=0.5, channel_ratio=0.25, concats=None, depth=6, csp_downsample_ratio=1, use_additional_stack=False, activation="swish", name=""
):
    # yolov7                                                        # yolov7_w6
    # 51: p5 512 ---+---------------------+-> 101: out2 512         # 47: p5 512 ---┬---------------------┬-> 113: out 512
    #               v [up 256 -> concat]  ^ [down 512 -> concat]    #               ↓ [up 384 -> concat]  ↑[down 512 -> concat]
    # 37: p4 1024 -> 63: p4p5 256 -------> 88: out1 256             # 37: p4 768 --- 59: p4p5 384 ------- 103: out 384
    #               v [up 128 -> concat]  ^ [down 256 -> concat]    #               ↓ [up 256 -> concat]  ↑[down 384 -> concat]
    # 24: p3 512 --> 75: p3p4p5 128 ------+--> 75: out0 128         # 28: p3 512 --- 71: p3p4p5 256 -- 93: out 256
    #                                                               #               ↓ [up 128 -> concat]  ↑[down 256 -> concat]
    #                                                               # 19: p2 256 --- 83: p2p3p4p5 128 -----┴-> 83: out 128
    # features: [p3, p4, p5]
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    hidden_channels = hidden.copy() if isinstance(hidden, list) else hidden
    upsamples = [features[-1]]
    p_name = "p{}_".format(len(features) + 2)
    # upsamples: [p5], features[:-1][::-1]: [p4, p3] -> [p5, p4p5, p3p4p5]
    for id, ii in enumerate(features[:-1][::-1]):
        cur_p_name = "p{}".format(len(features) + 1 - id)
        nn = conv_bn(ii, int(ii.shape[channel_axis] * channel_ratio), kernel_size=1, activation=activation, name=name + cur_p_name + "_down_")
        hidden_channel = hidden_channels.pop(0) if isinstance(hidden_channels, list) else hidden_channels
        p_name = cur_p_name + p_name
        nn = upsample_merge([nn, upsamples[-1]], hidden_channel, mid_ratio, concats, depth, use_additional_stack, activation=activation, name=name + p_name)
        upsamples.append(nn)

    downsamples = [upsamples[-1]]
    # downsamples: [p3p4p5], upsamples[:-1][::-1]: [p4p5, p5] -> [p3p4p5, p3p4p5 + p4p5, p3p4p5 + p4p5 + p5]
    for id, ii in enumerate(upsamples[:-1][::-1]):
        cur_name = name + "c3n{}_".format(id + 3)
        hidden_channel = hidden_channels.pop(0) if isinstance(hidden_channels, list) else hidden_channels
        cur_csp_downsample_ratio = csp_downsample_ratio.pop(0) if isinstance(csp_downsample_ratio, list) else csp_downsample_ratio
        nn = downsample_merge(
            [downsamples[-1], ii], hidden_channel, mid_ratio, concats, depth, cur_csp_downsample_ratio, use_additional_stack, activation, name=cur_name
        )
        downsamples.append(nn)
    return downsamples


""" YOLOV7Head, using Reparam Conv block """


def yolov7_head_single(inputs, filters, use_reparam_conv_head=True, num_classes=80, num_anchors=3, use_object_scores=True, activation="swish", name=""):
    if use_reparam_conv_head:
        # OREPA_3x3_RepConv
        rep_conv_3 = conv_bn(inputs, filters, 3, activation=None, name=name + "3x3_")
        rep_conv_1 = conv_bn(inputs, filters, 1, activation=None, name=name + "1x1_")
        nn = layers.Add()([rep_conv_3, rep_conv_1])
        nn = activation_by_name(nn, activation=activation, name=name)
    else:
        nn = conv_bn(inputs, filters, 3, activation=activation, name=name + "1_")

    ouput_classes = num_classes + (5 if use_object_scores else 4)  # num_anchors = 3, num_anchors * (80 + 5) = 255
    nn = layers.Conv2D(ouput_classes * num_anchors, kernel_size=1, name=name + "2_conv")(nn)
    nn = nn if image_data_format() == "channels_last" else layers.Permute([2, 3, 1])(nn)
    # return nn
    return layers.Reshape([-1, ouput_classes], name=name + "output_reshape")(nn)


def yolov7_head(
    inputs, use_reparam_conv_head=True, num_classes=80, num_anchors=3, use_object_scores=True, activation="swish", classifier_activation="sigmoid", name=""
):
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    outputs = []
    for id, input in enumerate(inputs):
        cur_name = name + "{}_".format(id + 1)
        filters = int(input.shape[channel_axis] * 2)
        out = yolov7_head_single(input, filters, use_reparam_conv_head, num_classes, num_anchors, use_object_scores, activation=activation, name=cur_name)
        outputs.append(out)
    # return outputs
    outputs = functional.concat(outputs, axis=1)
    return activation_by_name(outputs, classifier_activation, name="classifier_")


""" YOLOV7 models, almost same with yolor """


def YOLOV7(
    backbone=None,
    csp_channels=[64, 128, 256, 256],  # [YOLOV7Backbone parameters]
    stack_concats=[-1, -3, -5, -6],
    stack_depth=6,
    stack_out_ratio=1.0,
    use_additional_stack=False,
    stem_width=-1,  # -1 means using csp_channels[0] // 2
    stem_type="conv3",  # One of ["conv3", "focus", "conv1"], "focus" for YOLOV7_*6 models, "conv1" for YOLOV7_Tiny
    csp_downsample_ratios=[0, 0.5, 0.5, 0.5],
    spp_depth=2,
    fpn_hidden_channels=[256, 128, 256, 512],  # [FPN parameters]
    fpn_channel_ratio=0.25,
    fpn_stack_concats=None,
    fpn_stack_depth=-1,  # -1 for using same with stack_depth
    fpn_mid_ratio=0.5,
    fpn_csp_downsample_ratio=1,
    use_reparam_conv_head=True,
    features_pick=[-3, -2, -1],  # [Detector parameters]
    anchors_mode="yolor",
    num_anchors="auto",  # "auto" means: anchors_mode=="anchor_free" -> 1, anchors_mode=="yolor" -> 3, else 9
    use_object_scores="auto",  # "auto" means: True if anchors_mode=="anchor_free" or anchors_mode=="yolor", else False
    input_shape=(640, 640, 3),
    num_classes=80,
    activation="swish",
    classifier_activation="sigmoid",
    freeze_backbone=False,
    pretrained=None,
    model_name="yolov7",
    pyramid_levels_min=3,  # Init anchors for model prediction.
    anchor_scale="auto",  # Init anchors for model prediction. "auto" means 1 if (anchors_mode=="anchor_free" or anchors_mode=="yolor"), else 4
    rescale_mode="raw01",  # For decode predictions, raw01 means input value in range [0, 1].
    kwargs=None,  # Not using, recieving parameter
):
    # Regard input_shape as force using original shape if first element is None or -1,
    # else assume channel dimention is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)

    if backbone is None:
        # Save line width...
        csp_kwargs = {"out_features": features_pick, "spp_depth": spp_depth, "input_shape": input_shape, "activation": activation, "model_name": "backbone"}
        backbone = YOLOV7Backbone(
            csp_channels, stack_concats, stack_depth, stack_out_ratio, use_additional_stack, stem_width, stem_type, csp_downsample_ratios, **csp_kwargs
        )
        features = backbone.outputs
    else:
        if isinstance(features_pick[0], str):
            features = [backbone.get_layer(layer_name) for layer_name in features_pick]
        else:
            features = model_surgery.get_pyramide_feature_layers(backbone)
            features = [features[id] for id in features_pick]
        feature_names, features = model_surgery.align_pyramide_feature_output_by_image_data_format(features)
        print(">>>> features:", {ii: jj.shape for ii, jj in zip(feature_names, features)})

    backbone.trainable = False if freeze_backbone else True
    use_object_scores, num_anchors, anchor_scale = anchors_func.get_anchors_mode_parameters(anchors_mode, use_object_scores, num_anchors, anchor_scale)
    inputs = backbone.inputs[0]

    # Save line width...
    fpn_stack_depth = fpn_stack_depth if fpn_stack_depth > 0 else stack_depth
    fpn_kwargs = {"csp_downsample_ratio": fpn_csp_downsample_ratio, "use_additional_stack": use_additional_stack, "activation": activation, "name": "pafpn_"}
    fpn_features = path_aggregation_fpn(features, fpn_hidden_channels, fpn_mid_ratio, fpn_channel_ratio, fpn_stack_concats, fpn_stack_depth, **fpn_kwargs)

    outputs = yolov7_head(fpn_features, use_reparam_conv_head, num_classes, num_anchors, use_object_scores, activation, classifier_activation, name="head_")
    outputs = layers.Activation("linear", dtype="float32", name="outputs_fp32")(outputs)
    model = models.Model(inputs, outputs, name=model_name)
    reload_model_weights(model, PRETRAINED_DICT, "yolov7", pretrained)

    pyramid_levels = [pyramid_levels_min, pyramid_levels_min + len(features_pick) - 1]  # -> [3, 5]
    post_process = eval_func.DecodePredictions(backbone.input_shape[1:], pyramid_levels, anchors_mode, use_object_scores, anchor_scale)
    add_pre_post_process(model, rescale_mode=rescale_mode, post_process=post_process)
    return model


def YOLOV7_Tiny(
    input_shape=(416, 416, 3),
    freeze_backbone=False,
    num_classes=80,
    backbone=None,
    activation="leaky_relu/0.1",
    classifier_activation="sigmoid",
    pretrained="coco",
    **kwargs,
):
    # anchors_yolov7_tiny = np.array([[10,13, 16,30, 33,23], [30,61, 62,45, 59,119], [116,90, 156,198, 373,326]])
    # anchors_yolor = np.array([[12,16, 19,36, 40,28], [36,75, 76,55, 72,146], [142,110, 192,243, 459,401]])
    # anchors_yolov7_tiny == np.ceil((anchors_yolor * 416 / 512)).astype('int') [TODO]
    stem_type = "conv1"
    csp_channels = [32, 64, 128, 256]
    stack_concats = [-1, -2, -3, -4]
    stack_depth = 4
    stack_out_ratio = 0.5
    csp_downsample_ratios = [0, [0, 0], [0, 0], [0, 0]]  # First 0 for conv_bn downsmaple, others [0, 0] means maxpool
    spp_depth = 1

    fpn_hidden_channels = [64, 32, 64, 128]
    fpn_mid_ratio = 1.0
    fpn_channel_ratio = 0.5
    fpn_csp_downsample_ratio = [0, 0]  # [0, 0] means using conv_bn downsmaple
    use_reparam_conv_head = False
    return YOLOV7(**locals(), model_name=kwargs.pop("model_name", "yolov7_tiny"), **kwargs)


def YOLOV7_CSP(input_shape=(640, 640, 3), freeze_backbone=False, num_classes=80, backbone=None, classifier_activation="sigmoid", pretrained="coco", **kwargs):
    return YOLOV7(**locals(), model_name=kwargs.pop("model_name", "yolov7_csp"), **kwargs)


def YOLOV7_X(input_shape=(640, 640, 3), freeze_backbone=False, num_classes=80, backbone=None, classifier_activation="sigmoid", pretrained="coco", **kwargs):
    stack_concats = [-1, -3, -5, -7, -8]
    stack_depth = 8
    stem_width = 80

    fpn_stack_concats = [-1, -3, -5, -7, -8]
    fpn_mid_ratio = 1.0
    use_reparam_conv_head = False
    return YOLOV7(**locals(), model_name=kwargs.pop("model_name", "yolov7_x"), **kwargs)


def YOLOV7_W6(input_shape=(1280, 1280, 3), freeze_backbone=False, num_classes=80, backbone=None, classifier_activation="sigmoid", pretrained="coco", **kwargs):
    csp_channels = kwargs.pop("csp_channels", [64, 128, 256, 384, 512])
    features_pick = kwargs.pop("features_pick", [-4, -3, -2, -1])
    stem_type = kwargs.pop("stem_type", "focus")
    csp_downsample_ratios = kwargs.pop("csp_downsample_ratios", [128, 256, 512, 768, 1024])  # > 1 value means using conv_bn instead of csp_downsample
    stack_out_ratio = kwargs.pop("stack_out_ratio", 0.5)

    fpn_hidden_channels = kwargs.pop("fpn_hidden_channels", [384, 256, 128, 256, 384, 512])
    fpn_channel_ratio = kwargs.pop("fpn_channel_ratio", 0.5)
    fpn_csp_downsample_ratio = kwargs.pop("fpn_csp_downsample_ratio", 0)
    use_reparam_conv_head = kwargs.pop("use_reparam_conv_head", False)

    kwargs.pop("kwargs", None)  # From other YOLOV7_*6 models
    return YOLOV7(**locals(), model_name=kwargs.pop("model_name", "yolov7_w6"), **kwargs)


def YOLOV7_E6(input_shape=(1280, 1280, 3), freeze_backbone=False, num_classes=80, backbone=None, classifier_activation="sigmoid", pretrained="coco", **kwargs):
    stack_concats = [-1, -3, -5, -7, -8]
    stack_depth = 8
    stem_width = 80
    csp_downsample_ratios = [1, 1, 1, [1, 480 / 640], [1, 640 / 960]]  # different from YOLOV7_W6

    fpn_mid_ratio = 0.5
    fpn_csp_downsample_ratio = [1, [1, 240 / 320], [1, 320 / 480]]  # different from YOLOV7_W6

    kwargs.pop("kwargs", None)  # From YOLOV7_E6E / YOLOV7_D6
    return YOLOV7_W6(**locals(), model_name=kwargs.pop("model_name", "yolov7_e6"), **kwargs)


def YOLOV7_D6(input_shape=(1280, 1280, 3), freeze_backbone=False, num_classes=80, backbone=None, classifier_activation="sigmoid", pretrained="coco", **kwargs):
    stack_concats = [-1, -3, -5, -7, -9, -10]
    stack_depth = 10
    stem_width = 96
    return YOLOV7_E6(**locals(), model_name=kwargs.pop("model_name", "yolov7_d6"), **kwargs)


def YOLOV7_E6E(input_shape=(1280, 1280, 3), freeze_backbone=False, num_classes=80, backbone=None, classifier_activation="sigmoid", pretrained="coco", **kwargs):
    use_additional_stack = True
    return YOLOV7_E6(**locals(), model_name=kwargs.pop("model_name", "yolov7_e6e"), **kwargs)



from keras_cv_attention_models.gpvit.gpvit import GPViT, GPViT_L1, GPViT_L2, GPViT_L3, GPViT_L4, PureWeigths

__head_doc__ = """
Keras implementation of [Github ChenhongyiYang/GPViT](https://github.com/ChenhongyiYang/GPViT).
Paper [PDF 2212.06795 GPVIT: A HIGH RESOLUTION NON-HIERARCHICAL VISION TRANSFORMER WITH GROUP PROPAGATION](https://arxiv.org/pdf/2212.06795.pdf).
"""

__tail_doc__ = """  input_shape: it should have exactly 3 inputs channels, like `(224, 224, 3)`.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  activation: activation used in whole model, default `gelu`.
  drop_connect_rate: is used for [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382).
      Can be a constant value like `0.2`,
      or a tuple value like `(0, 0.2)` indicates the drop probability linearly changes from `0 --> 0.2` for `top --> bottom` layers.
      A higher value means a higher probability will drop the deep branch.
      or `0` to disable (default).
  layer_scale: layer scale init value, [Going deeper with Image Transformers](https://arxiv.org/abs/2103.17239).
      Default 0 for not using.
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
      Default is `None`.
  pretrained: one of `None` (random initialization) or 'imagenet21k-ft1k' (pre-training on ImageNet21k and fine-tuned ImageNet).
      Will try to download and load pre-trained model weights if not None.

Returns:
    A `keras.Model` instance.
"""

GPViT.__doc__ = __head_doc__ + """
Args:
  num_layers: number of transformer blocks.
  embed_dims: output channels for each stack.
  stem_depth: number of stem conv blocks.
  num_window_heads: number of heads for `window_lepe_attention_mlp_block` blocks.
  num_group_heads: number of heads for `group_attention` blocks.
  mlp_ratios: int value indicates expand ratio for mlp blocks hidden channel in each stack.
  window_size: number of `window_size` for `window_lepe_attention_mlp_block` blocks.
  group_attention_layer_ids: list of layer id for using `group_attention`, others will be `window_lepe_attention_mlp_block`.
  group_attention_layer_group_tokens: list of `num_group_token` for each block using `group_attention`.
  use_neck_attention_output: boolean value whether using `light_group_attention` before output block.
  model_name: string, model name.
""" + __tail_doc__ + """
Model architectures:
  | Model    | Params | FLOPs  | Input | Top1 Acc |
  | -------- | ------ | ------ | ----- | -------- |
  | GPViT_L1 | 9.59M  | 6.15G  | 224   | 80.5     |
  | GPViT_L2 | 24.2M  | 15.74G | 224   | 83.4     |
  | GPViT_L3 | 36.7M  | 23.54G | 224   | 84.1     |
  | GPViT_L4 | 75.5M  | 48.29G | 224   | 84.3     |
"""

GPViT_L1.__doc__ = __head_doc__ + """
Args:
""" + __tail_doc__

GPViT_L2.__doc__ = GPViT_L1.__doc__
GPViT_L3.__doc__ = GPViT_L1.__doc__
GPViT_L4.__doc__ = GPViT_L1.__doc__
